# from samplics.apis.sae.area_eblup import _log_likelihood
from samplics.types.errors import (
    CertaintyError,
    DimensionError,
    MethodError,
    ProbError,
    SamplicsError,
    SinglePSUError,
)
from samplics.types.options import PopParam, SelectMethod, SinglePSUEst, SizeMethod


__all__ = [
    "PopParam",
    "SinglePSUEst",
    "SelectMethod",
    "SizeMethod",
    "SamplicsError",
    "CertaintyError",
    "DimensionError",
    "SinglePSUError",
    "ProbError",
    "MethodError",
    "fit_eblup",
    "predict_eblup",
    # "_log_likelihood",
]



"""Sampling selection module

The module has one main class called *SampleSelection* which provides a number of random selection
methods and associated probability of selection. All the samping techniques implemented in this
modules are discussed in the following reference book: Cochran, W.G. (1977) [#c1977]_,
Kish, L. (1965) [#k1965]_, and Lohr, S.L. (2010) [#l2010]_. Furthermore, Brewer, K.R.W. and
Hanif, M. (1983) [#bh1983]_ provides comprehensive and detailed descriptions of these complex
sampling algorithms.

.. [#c1977] Cochran, W.G. (1977), *Sampling Techniques, 3rd edn.*, Jonh Wiley & Sons, Inc.
.. [#k1965] Kish, L. (1965), *Survey Sampling*, Jonh Wiley & Sons, Inc.
.. [#l2010] Lohr, S.L. (2010), *Sampling: Design and Analysis, 2nd edn.*, Cengage Learning, Inc.
.. [#bh1983] Brewer, K.R.W. and Hanif, M. (1983), *Sampling With Unequal Probabilities*,
   Springer-Verlag New York, Inc
"""

from __future__ import annotations

import math

from dataclasses import InitVar, dataclass, field
from typing import Optional, Union

import numpy as np
import pandas as pd

from samplics.utils.errors import CertaintyError, MethodError, ProbError
from samplics.utils.formats import data_to_dict, numpy_array, remove_nans, sample_units
from samplics.utils.types import (
    Array,
    DictStrBool,
    DictStrFloat,
    DictStrInt,
    DictStrNum,
    Number,
    SelectMethod,
)


@dataclass
class SampleSelection:

    # __slot__ = (
    #     "method",
    #     "strat",
    #     "wr",
    #     "pop_size",
    #     "samp_size",
    #     "samp_rate",
    #     "strata",
    #     "shuffle",
    #     "remove_nan",
    # )

    method: InitVar[SelectMethod] = field(init=True, default=False)
    strat: InitVar[bool] = field(init=True, default=False)
    wr: InitVar[bool] = field(init=True, default=False)

    pop_size: Union[DictStrNum, Number] = field(init=False, default=0)
    samp_size: Union[DictStrNum, Number] = field(init=False, default=0)
    samp_rate: Union[DictStrNum, Number] = field(init=False, default=0)
    strata: list = field(init=False, default=list)

    # sample: np.ndarray = field(init=False, default=None)
    # hits: np.ndarray = field(init=False, default=None)
    # probs: np.ndarray = field(init=False, default=None)zzzzzz

    shuffle: InitVar[bool] = field(init=True, default=False)
    remove_nan: InitVar[bool] = field(init=True, default=False)

    def __init__(
        self,
        method: SelectMethod,
        strat: bool = False,
        wr: bool = True,
    ) -> None:
        self.method: SelectMethod = method
        self.strat = strat
        self.fpc: Union[DictStrNum, Number] = {}
        self.wr = wr
        if method in (SelectMethod.srs_wr, SelectMethod.pps_wr):
            self.wr = True
        elif method in (
            SelectMethod.srs_wor,
            SelectMethod.pps_brewer,
            SelectMethod.pps_hv,
            SelectMethod.pps_murphy,
            SelectMethod.pps_rs,
            # SelectMethod.pps_sys,
        ):
            self.wr = False
        else:
            self.wr = wr

    @staticmethod
    def _to_dataframe(
        samp_unit: np.ndarray,
        stratum: np.ndarray,
        mos: np.ndarray,
        sample: np.ndarray,
        hits: np.ndarray,
        probs: np.ndarray,
    ) -> pd.DataFrame:

        df = pd.DataFrame(
            {
                "_samp_unit": samp_unit,
                "_stratum": stratum,
                "_mos": mos,
                "_sample": sample,
                "_hits": hits,
                "_probs": probs,
            }
        )
        df.reset_index(drop=True, inplace=True)

        if stratum.shape in ((), (0,)):
            df.drop(columns=["_stratum"], inplace=True)
        if mos.shape in ((), (0,)):
            df.drop(columns=["_mos"], inplace=True)

        return df

    @staticmethod
    def _calculate_samp_size(
        strat: bool,
        pop_size: Union[DictStrInt, int],
        samp_rate: Union[DictStrInt, int],
    ) -> Union[DictStrInt, int]:
        if strat:
            samp_size = {}
            for s in samp_rate:
                samp_size[s] = math.ceil(samp_rate[s] * pop_size[s])
            return samp_size
        else:
            return math.ceil(samp_rate * pop_size)

    @staticmethod
    def _calculate_samp_rate(
        strat: bool,
        pop_size: Union[DictStrInt, int],
        samp_size: Union[DictStrInt, int],
    ):
        if strat:
            samp_rate = {}
            for s in samp_size:
                samp_rate[s] = samp_size[s] / pop_size[s]
            return samp_rate
        else:
            return samp_size / pop_size

    @staticmethod
    def _calculate_fpc(
        strat: bool,
        pop_size: Union[DictStrInt, int],
        samp_size: Union[DictStrInt, int],
    ) -> None:

        if strat:
            fpc = {}
            for s in samp_size:
                fpc[s] = np.sqrt((pop_size[s] - samp_size[s]) / (pop_size[s] - 1))
            return fpc
        else:
            return np.sqrt((pop_size - samp_size) / (pop_size - 1))

    @staticmethod
    def _grs_select(
        probs: np.ndarray,
        samp_unit: np.ndarray,
        samp_size: Union[DictStrInt, int],
        stratum: np.ndarray,
        wr: Union[DictStrBool, bool],
    ) -> tuple[np.ndarray, np.ndarray]:

        sample = np.zeros(samp_unit.size).astype(bool)
        hits = np.zeros(samp_unit.size).astype(int)

        if stratum.shape in ((), (0,)):
            sampled_indices = np.random.choice(
                a=samp_unit.size,
                size=samp_size,
                replace=wr,
                p=probs / np.sum(probs),
            )
        else:
            sampled_indices_list = []
            for s in samp_size:
                units_s = stratum == s
                sampled_indices_s = np.random.choice(
                    a=samp_unit[units_s],
                    size=samp_size[s],
                    replace=wr,
                    p=probs[units_s] / np.sum(probs[units_s]),
                )
                sampled_indices_list.append(sampled_indices_s)
            sampled_indices = np.array([val for sublist in sampled_indices_list for val in sublist]).flatten()

        indices_s, hits_s = np.unique(sampled_indices, return_counts=True)
        sample[indices_s] = True
        hits[indices_s] = hits_s

        return sample, hits

    @staticmethod
    def _anycertainty(
        samp_size: Union[DictStrInt, int],
        stratum: Optional[np.ndarray],
        mos: np.ndarray,
    ) -> bool:

        certainty = False
        if stratum.shape not in ((), (0,)) and isinstance(samp_size, dict):
            for s in np.unique(stratum):
                stratum_units = stratum == s
                mos_s = mos[stratum_units]
                if (samp_size[s] * mos_s / np.sum(mos_s) >= 1).any():
                    certainty = True
                    break
        elif isinstance(samp_size, (int, float)):
            if (samp_size * mos / np.sum(mos) >= 1).any():
                certainty = True

        return certainty

    # SRS methods
    def _srs_inclusion_probs(
        self,
        samp_unit: np.ndarray,
        samp_size: Union[DictStrInt, int],
        stratum: np.ndarray,
    ) -> np.ndarray:

        number_units = samp_unit.size
        if stratum.shape not in ((), (0,)) and isinstance(samp_size, dict):
            incl_probs = np.zeros(number_units) * np.nan
            for s in np.unique(stratum):
                number_units_s = samp_unit[stratum == s].size
                incl_probs[stratum == s] = samp_size[s] / number_units_s
        elif isinstance(samp_size, (int, float)):
            number_units = samp_unit.size
            incl_probs = np.ones(number_units) * samp_size / number_units
        else:
            raise TypeError("samp_size has the wrong type!")

        return incl_probs

    # PPS methods
    def _pps_inclusion_probs(
        self,
        samp_unit: np.ndarray,
        samp_size: Union[DictStrInt, int],
        mos: np.ndarray,
        stratum: np.ndarray,
    ) -> np.ndarray:

        if isinstance(samp_size, dict):
            number_units = samp_unit.size
            incl_probs = np.zeros(number_units) * np.nan
            for s in np.unique(stratum):
                stratum_units = stratum == s
                mos_s = mos[stratum_units]
                incl_probs[stratum_units] = samp_size[s] * mos_s / np.sum(mos_s)
        else:
            incl_probs = samp_size * mos / np.sum(mos)

        return incl_probs

    @staticmethod
    def _pps_sys_select_core(
        samp_unit: np.ndarray,
        samp_size: int,
        mos: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray]:
        cumsize = np.append(0, np.cumsum(mos))
        samp_interval = cumsize[-1] / samp_size
        random_start = np.random.random_sample() * samp_interval
        random_picks = random_start + samp_interval * np.linspace(0, samp_size - 1, samp_size)

        hits = np.zeros(samp_unit.size).astype(int)
        for k in range(cumsize.size - 1):
            for ll in range(random_picks.size):
                if cumsize[k] < random_picks[ll] <= cumsize[k + 1]:
                    hits[k] += 1

        return np.asarray(hits >= 1), hits

    def _pps_sys_select(
        self,
        samp_unit: np.ndarray,
        samp_size: Union[DictStrInt, int],
        stratum: np.ndarray,
        mos: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray]:

        samp_unit = sample_units(samp_unit, unique=True)
        samp_size = data_to_dict(samp_size, self.strat, stratum)

        sample = hits = np.zeros(samp_unit.size).astype(int)

        if isinstance(samp_size, dict):
            for s in np.unique(stratum):
                stratum_units = stratum == s
                (
                    sample[stratum_units],
                    hits[stratum_units],
                ) = self._pps_sys_select_core(
                    samp_unit=samp_unit[stratum_units],
                    samp_size=samp_size[s],
                    mos=mos[stratum_units],
                )
        else:
            sample, hits = self._pps_sys_select_core(samp_unit=samp_unit, samp_size=samp_size, mos=mos)

        return sample, hits

    @staticmethod
    def _pps_hv_select_core(
        samp_unit: np.ndarray,
        samp_size: int,
        mos: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray]:

        pop_size = samp_unit.size
        all_indices = np.arange(pop_size)
        size_order = mos.argsort()
        all_indices_sorted = all_indices[size_order]
        mos_sorted = mos[size_order]
        probs_sorted = np.append(mos_sorted / np.sum(mos_sorted), 1 / samp_size)
        last_nplus1_probs = probs_sorted[range(-(samp_size + 1), 0)]
        diff_probs = np.ediff1d(last_nplus1_probs)
        s = np.sum(probs_sorted[0 : (pop_size - samp_size)])
        initial_probs_selection = (
            samp_size
            * diff_probs
            * (1 + np.linspace(1, samp_size, samp_size) * probs_sorted[pop_size - samp_size] / s)
        )
        probs_sorted = np.delete(probs_sorted, -1)
        selected_i = np.random.choice(np.arange(0, samp_size), size=1, p=initial_probs_selection)[0]
        sampled_indices = all_indices_sorted[selected_i + 1 : samp_size]

        notsampled_indices = np.delete(all_indices_sorted, sampled_indices)
        notsampled_probs = np.delete(probs_sorted, sampled_indices)
        p_denominator = (
            s
            + np.linspace(
                1,
                pop_size - samp_size + selected_i + 1,
                pop_size - samp_size + selected_i + 1,
            )
            * probs_sorted[pop_size - samp_size]
        )
        p_starts = notsampled_probs / p_denominator
        range_part2 = range(pop_size - samp_size, pop_size - samp_size + selected_i)
        p_starts[range_part2] = probs_sorted[pop_size - samp_size] / p_denominator[range_part2]
        p_starts_sum = np.cumsum(np.flip(p_starts)[range(p_starts.size - 1)])
        p_starts_sum = np.append(np.flip(p_starts_sum), 1)
        p_double_starts = p_starts / p_starts_sum
        p_double_starts[-1] = 0

        start_j = 0
        end_j = pop_size - samp_size + 1
        for ll in np.arange(selected_i + 1):
            sampling_space = range(start_j, end_j)
            p_double_space = p_double_starts[range(start_j, end_j)]
            p_double_space = 1 - (selected_i + 1 - ll) * np.append(0, p_double_space)
            p_double_space = np.delete(p_double_space, -1)
            a_j = (samp_size - ll + 1) * p_starts[range(start_j, end_j)] * np.cumprod(p_double_space)
            indice_j = np.random.choice(sampling_space, size=1, p=a_j / np.sum(a_j))[0]
            selected_j = notsampled_indices[indice_j]
            sampled_indices = np.append(sampled_indices, selected_j)
            start_j = indice_j + 1
            end_j += 1
        sample = hits = np.zeros(samp_unit.size).astype(int)
        sample[sampled_indices] = True
        hits[sampled_indices] = 1

        return sample, hits

    def _pps_hv_select(
        self,
        samp_unit: np.ndarray,
        samp_size: Union[DictStrInt, int],
        stratum: np.ndarray,
        mos: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray]:

        samp_unit = sample_units(samp_unit, unique=True)
        samp_size = data_to_dict(samp_size, self.strat, stratum)

        sample = hits = np.zeros(samp_unit.size).astype(int)

        if isinstance(samp_size, dict):
            for s in np.unique(stratum):
                stratum_units = stratum == s
                (
                    sample[stratum_units],
                    hits[stratum_units],
                ) = self._pps_hv_select_core(
                    samp_unit=samp_unit[stratum_units],
                    samp_size=samp_size[s],
                    mos=mos[stratum_units],
                )
        else:
            sample, hits = self._pps_hv_select_core(samp_unit=samp_unit, samp_size=samp_size, mos=mos)

        return sample, hits

    @staticmethod
    def _pps_brewer_select_core(
        samp_unit: np.ndarray,
        samp_size: int,
        mos: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray]:

        all_indices = np.arange(samp_unit.size)
        all_probs = mos / np.sum(mos)
        working_probs = all_probs * (1 - all_probs) / (1 - samp_size * all_probs)
        working_probs = working_probs / np.sum(working_probs)
        sampled_indices = np.random.choice(all_indices, 1, p=working_probs)
        sample = hits = np.zeros(samp_unit.size).astype(int)
        for s in np.arange(1, samp_size):
            remaining_indices = np.delete(all_indices, sampled_indices)
            remaining_probs = np.delete(all_probs, sampled_indices)
            remaining_probs = remaining_probs * (1 - remaining_probs) / (1 - (samp_size - s) * remaining_probs)
            remaining_probs = remaining_probs / sum(remaining_probs)
            current_selection = np.random.choice(remaining_indices, 1, p=remaining_probs)
            sampled_indices = np.append(sampled_indices, current_selection)

        sample[sampled_indices] = True
        hits[sampled_indices] = 1

        return sample, hits

    def _pps_brewer_select(
        self,
        samp_unit: np.ndarray,
        samp_size: Union[DictStrInt, int],
        stratum: np.ndarray,
        mos: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray]:

        samp_unit = sample_units(samp_unit, unique=True)
        samp_size = data_to_dict(samp_size, self.strat, stratum)

        sample = hits = np.zeros(samp_unit.size).astype(int)

        if isinstance(samp_size, dict):
            for s in np.unique(stratum):
                stratum_units = stratum == s
                (
                    sample[stratum_units],
                    hits[stratum_units],
                ) = self._pps_brewer_select_core(
                    samp_unit=samp_unit[stratum_units],
                    samp_size=samp_size[s],
                    mos=mos[stratum_units],
                )
        else:
            sample, hits = self._pps_brewer_select_core(samp_unit=samp_unit, samp_size=samp_size, mos=mos)

        return sample, hits

    @staticmethod
    def _pps_murphy_select_core(
        samp_unit: np.ndarray,
        samp_size: int,
        mos: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray]:

        if samp_size != 2:
            raise ValueError("For the Murphy's selection algorithm, sample size must be equal to 2")
        all_indices = np.arange(samp_unit.size)
        all_probs = mos / np.sum(mos)
        sampled_indices = np.random.choice(all_indices, 1, p=all_probs)
        remaining_indices = np.delete(all_indices, sampled_indices)
        remaining_probs = np.delete(all_probs, sampled_indices)
        remaining_probs = remaining_probs / (1 - all_probs[sampled_indices])
        current_selection = np.random.choice(remaining_indices, 1, p=remaining_probs)
        sampled_indices = np.append(sampled_indices, current_selection)

        sample = hits = np.zeros(samp_unit.size).astype(int)
        sample[sampled_indices] = True
        hits[sampled_indices] = 1

        return sample, hits

    def _pps_murphy_select(
        self,
        samp_unit: np.ndarray,
        samp_size: Union[DictStrInt, int],
        stratum: np.ndarray,
        mos: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray]:

        samp_unit = sample_units(samp_unit, unique=True)
        samp_size = data_to_dict(samp_size, self.strat, stratum)

        sample = hits = np.zeros(samp_unit.size).astype(int)

        if isinstance(samp_size, dict):
            for s in np.unique(stratum):
                stratum_units = stratum == s
                (
                    sample[stratum_units],
                    hits[stratum_units],
                ) = self._pps_murphy_select_core(
                    samp_unit=samp_unit[stratum_units],
                    samp_size=samp_size[s],
                    mos=mos[stratum_units],
                )
        else:
            sample, hits = self._pps_murphy_select_core(samp_unit=samp_unit, samp_size=samp_size, mos=mos)

        return sample, hits

    @staticmethod
    def _pps_rs_select_core(
        samp_unit: np.ndarray,
        samp_size: int,
        mos: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray]:

        all_indices = np.arange(samp_unit.shape[0])
        all_probs = mos / np.sum(mos)

        stop = False
        sample = hits = np.zeros(samp_unit.shape[0]).astype(int)
        sampled_indices = None
        while stop is not True:
            sampled_indices = np.random.choice(all_indices, 1, p=all_probs)
            remaining_indices = np.delete(all_indices, sampled_indices)
            remaining_probs = all_probs / (1 - samp_size * all_probs)
            remaining_probs = np.delete(remaining_probs, sampled_indices)
            remaining_probs = remaining_probs / np.sum(remaining_probs)
            remaining_sample = np.random.choice(remaining_indices, samp_size - 1, p=remaining_probs)
            sampled_indices, counts = np.unique(np.append(sampled_indices, remaining_sample), return_counts=True)
            if (counts == 1).all():
                stop = True

        sample[sampled_indices] = True
        hits[sampled_indices] = 1

        return sample, hits

    def _pps_rs_select(
        self,
        samp_unit: np.ndarray,
        samp_size: Union[DictStrInt, int],
        stratum: np.ndarray,
        mos: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray]:

        samp_unit = sample_units(samp_unit, unique=True)
        samp_size = data_to_dict(samp_size, self.strat, stratum)

        sample = hits = np.zeros(samp_unit.size).astype(int)

        if isinstance(samp_size, dict):
            for s in np.unique(stratum):
                stratum_units = stratum == s
                (
                    sample[stratum_units],
                    hits[stratum_units],
                ) = self._pps_rs_select_core(
                    samp_unit=samp_unit[stratum_units],
                    samp_size=samp_size[s],
                    mos=mos[stratum_units],
                )
        else:
            sample, hits = self._pps_rs_select_core(samp_unit=samp_unit, samp_size=samp_size, mos=mos)

        return sample, hits

    def _sys_inclusion_probs(
        self,
        samp_unit: np.ndarray,
        samp_size: Union[DictStrInt, int],
        stratum: np.ndarray,
        samp_rate: Union[DictStrFloat, float],
    ) -> np.ndarray:

        pass

    @staticmethod
    def _sys_selection_method(
        samp_unit: np.ndarray,
        samp_size: Optional[int],
        samp_rate: Optional[float],
    ) -> tuple[np.ndarray, np.ndarray]:

        if samp_size is not None and samp_rate is not None:
            raise AssertionError(
                """Both samp_size and samp_rate are provided.
                Only one of the two parameters should be specified."""
            )

        if samp_rate is None and samp_size is None:
            raise AssertionError("samp_size or samp_rate must be provided!")

        if samp_rate is not None:
            samp_size = round(samp_rate * samp_unit.size)
        samp_interval = int(samp_unit.size / samp_size)  # same as 1 / samp_rate

        random_start = np.random.choice(range(0, samp_interval))
        random_picks = random_start + samp_interval * np.linspace(0, samp_size - 1, samp_size).astype(int)
        hits = np.zeros(samp_unit.size).astype(int)
        hits[random_picks] = 1

        return hits == 1, hits

    def _sys_select(
        self,
        samp_unit: np.ndarray,
        samp_size: Union[DictStrInt, int],
        stratum: np.ndarray,
        samp_rate: Union[DictStrFloat, float],
    ) -> tuple[np.ndarray, np.ndarray]:

        sample = hits = np.zeros(samp_unit.size).astype(int)
        if self.strat and (isinstance(samp_size, dict) or isinstance(samp_rate, dict)):
            for s in np.unique(stratum):
                samp_size_s = None
                if samp_size is not None:
                    samp_size_s = samp_size[s]
                samp_rate_s = None
                if samp_rate is not None:
                    samp_rate_s = samp_rate[s]
                stratum_units = stratum == s
                (
                    sample[stratum_units],
                    hits[stratum_units],
                ) = self._sys_selection_method(samp_unit[stratum_units], samp_size_s, samp_rate_s)
        elif isinstance(samp_size, int) or isinstance(samp_rate, float):
            samp_size_n = None if samp_size is None else samp_size
            samp_rate_n = None if samp_rate is None else samp_rate
            sample, hits = self._sys_selection_method(samp_unit, samp_size_n, samp_rate_n)

        return sample, hits

    def inclusion_probs(
        self,
        samp_unit: np.ndarray,
        samp_size: Optional[Union[DictStrInt, int]] = None,
        stratum: Optional[np.ndarray] = None,
        mos: Optional[np.ndarray] = None,
        samp_rate: Optional[Union[DictStrFloat, float]] = None,
    ) -> np.ndarray:

        samp_unit = sample_units(samp_unit, unique=True)

        samp_size_temp: Union[DictStrInt, int]
        if stratum is not None:
            stratum = numpy_array(stratum)
            if isinstance(samp_size, (int, float)):
                strata = np.unique(stratum)
                samp_size_temp = dict(zip(strata, np.repeat(int(samp_size), strata.shape[0])))
            elif isinstance(samp_size, dict):
                samp_size_temp = samp_size.copy()
            else:
                raise TypeError("samp_size or samp_rate has the wrong type")
        else:
            if isinstance(samp_size, (int, float)):
                samp_size_temp = int(samp_size)
            else:
                raise TypeError("samp_size or samp_rate has the wrong type")

        mos = numpy_array(mos) if mos is not None else np.ones(samp_unit.shape[0])

        samp_size_temp = data_to_dict(samp_size_temp, self.strat, stratum)

        if self.method in (SelectMethod.srs_wr, SelectMethod.srs_wor):
            incl_probs = self._srs_inclusion_probs(samp_unit, samp_size, stratum)
        elif self.method in (
            SelectMethod.pps_sys,
            SelectMethod.pps_hv,
            SelectMethod.pps_brewer,
            SelectMethod.pps_murphy,
            SelectMethod.pps_rs,
        ):
            if self._anycertainty(samp_size_temp, stratum, mos):
                raise AssertionError("Some clusters are certainties.")
            incl_probs = self._pps_inclusion_probs(samp_unit, samp_size_temp, mos, stratum)
        elif self.method == SelectMethod.sys:
            incl_probs = self._sys_inclusion_probs(samp_unit, samp_size_temp, stratum, samp_rate)
        else:
            raise ValueError("method not valid!")

        return incl_probs

    def joint_inclusion_probs(self) -> None:
        pass

    def select(
        self,
        samp_unit: Array,
        samp_size: Optional[Union[DictStrInt, int]] = None,
        stratum: Optional[Array] = None,
        mos: Optional[Array] = None,
        samp_rate: Optional[Union[DictStrFloat, float]] = None,
        probs: Optional[Array] = None,
        shuffle: bool = False,
        to_dataframe: bool = False,
        sample_only: bool = False,
        remove_nan: bool = False,
    ) -> Union[tuple[np.ndarray, np.ndarray, np.ndarray], pd.DataFrame]:

        if samp_size is not None and samp_rate is not None:
            raise AssertionError(
                """Both samp_size and samp_rate are provided.
                Only one of the two parameters should be specified."""
            )

        if self.strat and stratum is None:
            raise AssertionError("Stratum must be provided for stratified samples!")

        _samp_unit = numpy_array(samp_unit)
        _stratum = numpy_array(stratum)  # if stratum is not None else None
        _samp_unit = sample_units(_samp_unit, unique=True)
        if self.strat:
            self.strata = np.unique(_stratum).tolist()
            strata, nobs = np.unique(_stratum, return_counts=True)
            self.pop_size = dict(zip(strata, nobs))
        else:
            self.strata = []
            self.pop_size = _samp_unit.shape[0]
        if samp_rate is None:
            self.samp_size = data_to_dict(data=samp_size, strat=self.strat, stratum=_stratum)
            self.samp_rate = self._calculate_samp_rate(
                strat=self.strat, pop_size=self.pop_size, samp_size=self.samp_size
            )
        else:
            self.samp_rate = data_to_dict(data=samp_rate, strat=self.strat, stratum=_stratum)
            self.samp_size = self._calculate_samp_size(
                strat=self.strat, pop_size=self.pop_size, samp_rate=self.samp_rate
            )

        self.fpc = self._calculate_fpc(strat=self.strat, pop_size=self.pop_size, samp_size=self.samp_size)

        if self.method == SelectMethod.grs:
            _probs = numpy_array(probs)
            if _probs.shape in ((), (0,)) or (_probs < 0).any() or (_probs > 1).any():
                raise ProbError("Probabilities must be between 0 and 1!")
        else:
            _probs = np.array(None)

        _mos = numpy_array(mos)
        if _mos.shape in ((), (0,)) and self.method in (
            SelectMethod.pps_brewer,
            SelectMethod.pps_hv,
            SelectMethod.pps_murphy,
            SelectMethod.pps_rs,
            SelectMethod.pps_sys,
        ):
            raise MethodError("Measure of size (MOS) not provided!")
        elif _mos.shape not in ((), (0,)) and self.method in (
            SelectMethod.pps_brewer,
            SelectMethod.pps_hv,
            SelectMethod.pps_murphy,
            SelectMethod.pps_rs,
            SelectMethod.pps_sys,
        ):
            if self._anycertainty(samp_size=self.samp_size, stratum=_stratum, mos=_mos):
                raise CertaintyError("Some clusters are certainties.")

        _samp_ids = np.linspace(start=0, stop=_samp_unit.shape[0] - 1, num=_samp_unit.shape[0], dtype="int")

        if remove_nan:
            items_to_keep = remove_nans(_samp_ids.shape[0], _samp_ids, _stratum, _mos, _probs)
            _samp_ids = _samp_ids[items_to_keep]
            _stratum = _stratum[items_to_keep]
            _mos = _mos[items_to_keep]
            _probs = _probs[items_to_keep]

        suffled_order = None
        if shuffle and self.method in (SelectMethod.sys, SelectMethod.pps_sys):
            suffled_order = np.linspace(start=0, stop=self.pop_size - 1, num=self.pop_size, dtype="int")
            np.random.shuffle(suffled_order)
            _samp_unit = _samp_unit[suffled_order]
            if _stratum.shape not in ((), (0,)):
                _stratum = _stratum[suffled_order]
            if self.method == SelectMethod.pps_sys and _mos.shape not in ((), (0,)):
                _mos = _mos[suffled_order]

        if self.method in (SelectMethod.srs_wr, SelectMethod.srs_wor):
            _probs = self._srs_inclusion_probs(samp_unit=_samp_ids, samp_size=self.samp_size, stratum=_stratum)
            sample, hits = self._grs_select(
                probs=_probs,
                samp_unit=_samp_ids,
                samp_size=self.samp_size,
                stratum=_stratum,
                wr=self.wr,
            )
        elif self.method == SelectMethod.sys:
            sample, hits = self._sys_select(
                samp_unit=_samp_ids,
                samp_size=None,
                samp_rate=self.samp_rate,
                stratum=_stratum,
            )
        elif self.method == SelectMethod.pps_wr:
            _probs = self._pps_inclusion_probs(
                samp_unit=_samp_ids,
                samp_size=self.samp_size,
                samp_rate=self.samp_rate,
                mos=_mos,
                stratum=_stratum,
            )
            sample, hits = self._grs_select(
                probs=_probs / np.sum(probs),
                samp_unit=_samp_ids,
                samp_size=self.samp_size,
                stratum=_stratum,
                wr=self.wr,
            )
        elif self.method == SelectMethod.pps_sys:
            _probs = self.inclusion_probs(
                samp_unit=_samp_ids,
                samp_size=self.samp_size,
                samp_rate=self.samp_rate,
                stratum=_stratum,
                mos=_mos,
            )
            sample, hits = self._pps_sys_select(
                samp_unit=_samp_ids,
                samp_size=self.samp_size,
                stratum=_stratum,
                mos=_mos,
            )
        elif self.method == SelectMethod.pps_brewer:
            _probs = self.inclusion_probs(
                samp_unit=_samp_ids,
                samp_size=self.samp_size,
                samp_rate=self.samp_rate,
                stratum=_stratum,
                mos=_mos,
            )
            sample, hits = self._pps_brewer_select(
                samp_unit=_samp_ids,
                samp_size=self.samp_size,
                stratum=_stratum,
                mos=_mos,
            )
        elif self.method == SelectMethod.pps_hv:
            _probs = self.inclusion_probs(
                samp_unit=_samp_ids,
                samp_size=self.samp_size,
                samp_rate=self.samp_rate,
                stratum=_stratum,
                mos=_mos,
            )
            sample, hits = self._pps_hv_select(
                samp_unit=_samp_ids,
                samp_size=self.samp_size,
                stratum=_stratum,
                mos=_mos,
            )
        elif self.method == SelectMethod.pps_murphy:
            _probs = self.inclusion_probs(
                samp_unit=_samp_ids,
                samp_size=self.samp_size,
                samp_rate=self.samp_rate,
                stratum=_stratum,
                mos=_mos,
            )
            sample, hits = self._pps_murphy_select(
                samp_unit=_samp_ids,
                samp_size=self.samp_size,
                stratum=_stratum,
                mos=_mos,
            )
        elif self.method == SelectMethod.pps_rs:
            _probs = self.inclusion_probs(
                samp_unit=_samp_ids,
                samp_size=self.samp_size,
                samp_rate=self.samp_rate,
                stratum=_stratum,
                mos=_mos,
            )
            sample, hits = self._pps_rs_select(
                samp_unit=_samp_ids,
                samp_size=self.samp_size,
                stratum=_stratum,
                mos=_mos,
            )
        elif self.method == SelectMethod.grs:
            sample, hits = self._grs_select(
                probs=_probs,
                samp_unit=_samp_ids,
                samp_size=self.samp_size,
                stratum=_stratum,
                wr=self.wr,
            )
        else:
            raise ValueError("sampling method not implemented!")

        if shuffle:
            sample = sample[suffled_order]
            hits = hits[suffled_order]

        if to_dataframe and sample_only:
            frame = self._to_dataframe(
                samp_unit=_samp_unit,
                stratum=_stratum,
                mos=_mos,
                sample=sample,
                hits=hits,
                probs=_probs,
            )
            return frame.loc[frame["_sample"] == 1].reset_index(drop=True)
        elif to_dataframe and not sample_only:
            frame = self._to_dataframe(
                samp_unit=_samp_unit,
                stratum=_stratum,
                mos=_mos,
                sample=sample,
                hits=hits,
                probs=_probs,
            )
            return frame
        elif not to_dataframe and sample_only and probs is not None:
            return sample[sample], hits[sample], probs[sample]
        else:
            return sample, hits, _probs



"""This module provides basic functions used across multiple classes and modules.

Functions:
    | *sumby()* sums numpy arrays by some grouping.
    | *averageby()* average numpy arrays by some grouping.
    | *transform()* implements the Boxcox tranformation.
    | *skewness()* computes the Pearson's moment coefficient of skewness (assymetry measure).
    | *kurtosis()* computes the fourth standardized moment, a measure of assymetry.
    | *plot_skewness()* and *plot_kurtosis()* visualise the skewness and kurtosis coefficients
respectively.

"""

from __future__ import annotations

from typing import Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from samplics.utils.formats import numpy_array
from samplics.utils.types import Array, Number


def set_variables_names(
    vars: Array, varnames: Optional[Union[str, list[str]]], prefix: str
) -> Union[str, list[str]]:

    if varnames is None:
        if isinstance(vars, pd.DataFrame):
            return list(vars.columns)
        elif isinstance(vars, pd.Series) and vars.name is not None:
            return [vars.name]
        else:
            if isinstance(vars, (pd.Series, np.ndarray)):
                if len(vars.shape) == 2:
                    return [prefix + "_" + str(k) for k in range(1, vars.shape[1] + 1)]
                else:
                    return [
                        prefix + "_" + str(k) for k in range(1, len(vars.shape) + 1)
                    ]
            elif isinstance(vars, (tuple, list)):
                return [prefix + "_" + str(k) for k in range(1, len(vars) + 1)]
            else:
                raise TypeError("vars should be an array-like")
    else:
        return varnames


def sumby(
    group: np.ndarray, y: np.ndarray
) -> np.ndarray:  # Could use pd.grouby().sum(), may scale better
    """Sums the numpy array by group.

    Args:
        group (np.ndarray): provides the group associated to each observation.
        y (np.ndarray): variable of interest for the summation.

    Returns:
        np.ndarray: the sums by group.
    """

    groups = np.unique(group)
    sums = np.zeros(groups.size)
    for k, gr in enumerate(groups):
        sums[k] = np.sum(y[group == gr])

    return sums


def averageby(
    group: np.ndarray, y: np.ndarray
) -> np.ndarray:  # Could use pd.grouby().sum(), may scale better
    """[summary]

    Args:
        group (np.ndarray): provides the group associated to each observation.
        y (np.ndarray): variable of interest for the averaging.

    Returns:
        np.ndarray: the averages by group.
    """

    groups = np.unique(group)
    means = np.zeros(groups.size)
    for k, gr in enumerate(groups):
        means[k] = np.mean(y[group == gr])

    return means


def transform(
    y: np.ndarray,
    llambda: Optional[Number] = None,
    constant: Optional[Number] = None,
    inverse: bool = True,
) -> np.ndarray:
    """Transforms the variable of interest using the Boxcox method or its inverse.

    Args:
        y (np.ndarray): variable of interest.
        llambda (Optional[Number], optional): Boxcox parameter lambda. A value of zero indicates
        a logarithm transformation. Defaults to None.
        constant (float, optional): An additive term to ensure that the logarithm function is
        applied to positive values. Defaults to 0.0.
        inverse (bool, optional): indicates if the Boxcox transformation is applied or its
        inverse. Defaults to True.

    Returns:
        np.ndarray: [description]
    """
    if llambda is None:
        return y
    elif llambda == 0.0:
        if constant is None:
            constant = 0.0
        if inverse:
            return np.asarray(np.exp(y) - constant)
        else:
            if ((y + constant) <= 0).any():
                raise ValueError("log function not defined for negative numbers")
            else:
                return np.asarray(np.log(y + constant))
    elif llambda != 0.0:
        if inverse:
            if ((1 + y * llambda) <= 0).any():
                raise ValueError("log function not defined for negative numbers")
            return np.asarray(np.exp(np.log(1 + y * llambda) / llambda))
        else:
            return np.asarray(np.power(y, llambda) / llambda)
    else:
        raise AssertionError


def skewness(y: Array, type: int = 1) -> float:

    if type not in (1, 2, 3):
        raise AssertionError("Parameter type must be 1, 2 or 3.")

    y = numpy_array(y)
    skewness = float(np.mean((y - np.mean(y)) ** 3) / np.std(y) ** 3)
    if type == 2:
        n = y.shape[0]
        if n <= 3:
            raise ValueError("For type 2, y must be of size 3 or more.")
        else:
            skewness = skewness * np.sqrt(n * (n - 1)) / (n - 2)
    elif type == 3:
        n = y.shape[0]
        skewness = skewness * np.power(1 - 1 / n, 3 / 2)

    return skewness


def kurtosis(y: Array, type: int = 1) -> float:

    if type not in (1, 2, 3):
        raise AssertionError("Parameter type must be 1, 2 or 3.")

    y = numpy_array(y)
    kurtosis = float(np.mean((y - np.mean(y)) ** 4) / np.std(y) ** 4)
    if type == 1:
        kurtosis = kurtosis - 3
    elif type == 2:
        n = y.shape[0]
        if n <= 4:
            raise ValueError("For type 2, y must be of size 4 or more.")
        kurtosis = ((n + 1) * (kurtosis - 3) + 6) * (n - 1) / ((n - 2) * (n - 3))
    elif type == 3:
        n = y.shape[0]
        kurtosis = kurtosis * np.power(1 - 1 / n, 2) - 3

    return kurtosis


def _plot_measure(
    y: np.ndarray,
    coef_min: Number = -5,
    coef_max: Number = 5,
    nb_points: int = 100,
    measure: str = "skewness",
    block: bool = True,
) -> None:
    y = numpy_array(y)
    lambda_range = np.linspace(coef_min, coef_max, num=nb_points)
    coefs = np.zeros(lambda_range.size)
    measure_loc = None
    for k, ll in enumerate(lambda_range):
        if (1 + y * ll < 0).any():
            break
        y_ll = transform(y, ll)
        if measure.lower() == "skewness":
            coefs[k] = skewness(y_ll)
            measure_loc = "lower right"
        elif measure.lower() == "kurtosis":
            coefs[k] = kurtosis(y_ll)
            measure_loc = "upper right"
        else:
            raise ValueError("measure type not valid!")

    normality = np.abs(coefs) < 2.0

    p1 = plt.scatter(
        lambda_range[normality],
        coefs[normality],
        marker="D",
        c="green",
        s=25,
        alpha=0.3,
    )
    p2 = plt.scatter(
        lambda_range[~normality],
        coefs[~normality],
        c="red",
        s=10,
        alpha=0.6,
        edgecolors="none",
    )
    plt.axhline(0, color="blue", linestyle="--")
    plt.title(f"{measure.title()} by BoxCox lambda")
    plt.ylabel(f"{measure.title()}")
    plt.xlabel("Lambda (coefs)")
    plt.legend(
        (p1, p2),
        ("Normality zone", "Non-normality zone"),
        loc=measure_loc,
    )
    plt.show(block=block)


def plot_skewness(
    y: np.ndarray,
    coef_min: Number = -5,
    coef_max: Number = 5,
    nb_points: int = 100,
    block: bool = True,
) -> None:
    """Plots a scatter plot of skewness coefficients and the lambda coefficients of the Boxcox
    transformation. The plot helps identify the range of lambda values to minimize the
    skewness coefficient.

    Args:
        y (np.ndarray): variable of interest.
        coef_min (Number, optional): lowest value for the horizontal axis. Defaults to -5.
        coef_max (Number, optional): highest value for the horizontal axis. Defaults to 5.
        nb_points (int, optional): number of points in the plot. Defaults to 100.
    """
    _plot_measure(
        y=y,
        coef_min=coef_min,
        coef_max=coef_max,
        nb_points=nb_points,
        measure="skewness",
        block=block,
    )


def plot_kurtosis(
    y: np.ndarray,
    coef_min: Number = -5,
    coef_max: Number = 5,
    nb_points: int = 100,
    block: bool = True,
) -> None:
    """Plots a scatter plot of skewness coefficients and the lambda coefficients of the Boxcox
    transformation. The plot helps identify the range of lambda values to minimize the
    kurtosis coefficient.

    Args:
        y (np.ndarray): variable of interest.
        coef_min (Number, optional): lowest value for the horizontal axis. Defaults to -5.
        coef_max (Number, optional): highest value for the horizontal axis. Defaults to 5.
        nb_points (int, optional): number of points in the plot. Defaults to 100.
    """
    _plot_measure(
        y=y,
        coef_min=coef_min,
        coef_max=coef_max,
        nb_points=nb_points,
        measure="kurtosis",
        block=block,
    )


def get_single_psu_strata(stratum: Array, psu: Array) -> Optional(np.ndarray):
    stratum = numpy_array(stratum)
    psu = numpy_array(psu)

    if psu.shape in (
        (),
        (0,),
    ):  # psu is None will not work because psu is an np.ndarray
        strata_ids, psu_counts = np.unique(stratum, return_counts=True)
    else:
        strata_ids, psu_counts = np.unique(
            np.unique((stratum, psu), axis=1)[0, :], return_counts=True
        )
    # breakpoint()

    single_psus = psu_counts == 1
    if single_psus.sum() == 0:
        return None
    else:
        return strata_ids[single_psus]



"""EBLUP Area Model

"""


import numpy as np
import polars as pl

from samplics.types import (
    AuxVars,
    DictStrNum,
    DirectEst,
    EblupEst,
    FitMethod,
    FitStats,
    Number,
)


def _predict_eblup(
    x: AuxVars,
    fit_eblup: FitStats,
    y: DirectEst,
    intercept: bool = True,  # if True, it adds an intercept of 1
    b_const: DictStrNum | Number = 1.0,
) -> EblupEst:
    area = y.to_numpy(keep_vars="__domain").flatten()

    if isinstance(b_const, (int, float)):
        b_const = dict(zip(area, np.ones(area.size) * b_const))

    sigme2_e = {}
    for d in fit_eblup.err_stderr:
        sigme2_e[d] = fit_eblup.err_stderr[d] ** 2

    est, mse, mse1, mse2, g1, g2, g3, g3_star = _eblup_estimates(
        method=fit_eblup.method,
        yhat=y.est,
        auxvars=x,
        area=area,
        beta=np.array(fit_eblup.fe_est),
        sigma2_e=sigme2_e,
        sigma2_v=fit_eblup.re_stderr**2,
        sigma2_v_cov=fit_eblup.re_stderr_var,
        intercept=intercept,
        b_const=b_const,
    )

    return EblupEst(pred=est, fit_stats=fit_eblup, domain=None, mse=mse, mse_boot=None, mse_jkn=None)


def _eblup_estimates(
    method: FitMethod,
    yhat: dict,
    auxvars: np.ndarray,
    beta: np.ndarray,
    area: np.ndarray,
    sigma2_e: dict,
    sigma2_v: Number,
    sigma2_v_cov: Number,
    intercept: bool,
    b_const: dict,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray,]:
    m = len(yhat)
    b_const_vec = np.array(list(b_const.values()))
    v_i = np.array(list(sigma2_e.values())) + sigma2_v * (b_const_vec**2)
    v_inv = np.diag(1 / v_i)
    G = np.diag(np.ones(m) * sigma2_v)
    Z = np.diag(b_const_vec)
    b = (G @ np.transpose(Z)) @ v_inv

    if intercept:
        x = np.insert(auxvars.to_numpy(drop_vars=["__record_id", "__domain"]), 0, 1, axis=1)  # add the intercept
    else:
        x = auxvars.to_numpy(drop_vars=["__record_id", "__domain"])

    d = np.transpose(x - np.transpose(b) @ x)
    x_vinv_x = np.transpose(x) @ v_inv @ x

    g2_term = np.linalg.pinv(x_vinv_x)

    b_term_ml1 = g2_term  # np.linalg.inv(x_vinv_x)
    b_term_ml2_diag = (b_const_vec**2) / (v_i**2)
    b_term_ml2 = np.matmul(np.matmul(np.transpose(x), np.diag(b_term_ml2_diag)), x)
    b_term_ml = float(np.trace(np.matmul(b_term_ml1, b_term_ml2)))

    estimates = {}  # np.zeros(m) * np.nan
    g1 = {}  # np.zeros(m) * np.nan
    g2 = {}  # np.zeros(m) * np.nan
    g3 = {}  # np.zeros(m) * np.nan
    g3_star = {}  # np.zeros(m) * np.nan

    g1_partial = {}  # np.zeros(m) * np.nan

    sum_inv_vi2 = np.sum(1 / (v_i**2))
    b_sigma2_v = 0.0
    if method == FitMethod.reml:
        g3_scale = 2.0 / sum_inv_vi2
    elif method == FitMethod.ml:
        b_sigma2_v = -(1.0 / 2.0 * sigma2_v_cov) * b_term_ml
        g3_scale = 2.0 / sum_inv_vi2
    elif method == FitMethod.fh:
        sum_vi = np.sum((1 / v_i))
        b_sigma2_v = 2.0 * (m * sum_inv_vi2 - sum_vi**2) / (sum_vi**3)
        g3_scale = 2.0 * m / sum_vi**2
    else:
        g3_scale = 0.0

    mse = {}
    mse1_area_specific = {}
    mse2_area_specific = {}

    for d in area:
        b_d = b_const[d]
        phi_d = sigma2_e[d]

        if intercept:
            X_d = np.insert(pl.from_dict(auxvars.x[d]).to_numpy(), 0, 1, axis=1)
        else:
            X_d = pl.from_dict(auxvars.x[d]).to_numpy()

        yhat_d = yhat[d]
        mu_d = X_d @ beta
        resid_d = yhat_d - mu_d
        variance_d = sigma2_v * (b_d**2) + phi_d
        gamma_d = sigma2_v * (b_d**2) / variance_d
        estimates[d] = (gamma_d * yhat_d + (1 - gamma_d) * mu_d)[0]
        g1[d] = gamma_d * phi_d
        g2_term_d = ((X_d @ g2_term) @ np.transpose(X_d))[0][0]
        g2[d] = ((1 - gamma_d) ** 2) * (g2_term_d)
        g3[d] = ((1 - gamma_d) ** 2) * g3_scale / variance_d
        g3_star[d] = ((g3[d] / variance_d) * (resid_d**2))[0]
        g1_partial[d] = (b_d**2) * ((1 - gamma_d) ** 2) * b_sigma2_v

        if method == FitMethod.reml:
            mse[d] = g1[d] + g2[d] + 2 * g3[d]
            mse1_area_specific[d] = g1[d] + g2[d] + 2 * g3_star[d]
            mse2_area_specific[d] = g1[d] + g2[d] + g3[d] + g3_star[d]
        elif method in (FitMethod.fh, FitMethod.ml):
            mse[d] = g1 - g1_partial[d] + g2[d] + 2 * g3[d]
            mse1_area_specific[d] = g1[d] - g1_partial[d] + g2[d] + 2 * g3_star[d]
            mse2_area_specific[d] = g1[d] - g1_partial[d] + g2[d] + g3[d] + g3_star[d]

    return (
        estimates,
        mse,
        mse1_area_specific,
        mse2_area_specific,
        g1,
        g2,
        g3,
        g3_star,
    )



"""Runs checks on the data to help capture some errors faster and provides better error messages.

Functions:
    | *assert_probabilities()* ensures that probability values are between 0 and 1.
    | *assert_weights()* ensures that sample weights are not negatives.
    | *assert_not_unique()* return an assertion error if the array has non unique values.
    | *assert_response_status()* checks that the response values are in ("in", "rr", "nr", "uk").
    | *assert_brr_number_psus()* checks that the number of psus is a multiple of 2.
"""
from __future__ import annotations

from typing import Iterable, Optional, Union

import numpy as np
import pandas as pd

from samplics.utils import formats
from samplics.utils.types import Array, Number, StringNumber


def assert_probabilities(**kargs: Union[Number, Iterable]) -> None:

    err_msg = "Probabilities must be between 0 and 1, inclusively!"
    for k in kargs:
        if not assert_in_range(0, 1, x=kargs[k]):
            raise ValueError(err_msg)


def assert_proportions(**kargs: Union[Number, Iterable]) -> None:

    err_msg = "Proportions must be between 0 and 1, inclusively!"
    for k in kargs:
        if not assert_in_range(0, 1, x=kargs[k]):
            raise ValueError(err_msg)


def assert_in_range(low: Number, high: Number, x: Union[Number, Iterable]) -> bool:
    if isinstance(x, (int, float)):
        if x > high or x < low:
            return False
    elif isinstance(x, (np.ndarray, pd.Series)):
        if (x > high).any() or (x < low).any():
            return False
    elif isinstance(x, Iterable):
        for i in x:
            if isinstance(x, dict):
                if x[i] > high or x[i] < low:
                    return False
            elif i > high or i < low:
                return False

    return True


def assert_weights(weights: Array) -> None:
    weights = formats.numpy_array(weights)
    if (weights < 0).any():
        raise ValueError("Sample weights must be positive values")


def assert_not_unique(array_unique_values: Array) -> None:
    if np.unique(array_unique_values).size != len(array_unique_values):
        raise AssertionError(
            "The array must contains unique values identifying all the units in the sampling frame."
        )


def assert_response_status(
    response_status: Union[str, np.ndarray],
    response_dict: Optional[dict[str, StringNumber]],
) -> None:
    if response_status is None:
        raise AssertionError("response_status is not provided")
    elif (
        not np.isin(response_status, ("in", "rr", "nr", "uk")).all()
        and response_dict is None
    ):
        raise AssertionError(
            "The response status must only contains values in ('in', 'rr', 'nr', 'uk') or the mapping should be provided using response_dict parameter"
        )
    elif "rr" not in response_status and response_dict is None:
        raise AssertionError("The response status must at least contains rr!")
    elif isinstance(response_dict, dict):
        # resp_keys = list(response_dict.keys())
        resp_keys = [x.lower() for x in response_dict]
        if not np.isin(resp_keys, ("in", "rr", "nr", "uk")).all():
            raise AssertionError("Response mapping dictionnary has unexpected value(s)")


def assert_brr_number_psus(psu: np.ndarray) -> None:
    if psu.size % 2 != 0:
        raise AssertionError(
            "For the BRR method, the number of PSUs must be a multiple of two."
        )



"""Provides the custom types used throughout the modules.
"""

from enum import Enum, unique
from typing import Dict

import numpy as np
import pandas as pd
import polars as pl


DF = pl.DataFrame | pd.DataFrame
Array = np.ndarray | pd.Series | pl.Series | list | tuple
Series = pd.Series | pl.Series | list | tuple

Number = float | int
StringNumber = str | float | int

DictStrNum = Dict[StringNumber, Number]
DictStrInt = Dict[StringNumber, int]
DictStrFloat = Dict[StringNumber, float]
DictStrBool = Dict[StringNumber, bool]



@unique
class FitMethod(Enum):
    ols= "OLS"
    wls= "WLS"
    fh = "FH"
    ml= "ML"
    reml= "REML"

# Population parameters
@unique
class PopParam(Enum):
    count = "Count"
    mean = "Mean"
    total = "Total"
    prop = "Proportion"
    ratio = "Ratio"


@unique
class RepMethod(Enum):
    jackknife  = "Jackknife"
    bootstrap = "Bootstrap"
    brr = "BRR"
    

# Methods for sample size
@unique
class SizeMethod(Enum):
    wald = "Wald"
    fleiss = "Fleiss"


@unique
class SelectMethod(Enum):
    srs_wr = "SRS with replacement"
    srs_wor = "SRS without replacement"
    sys = "Systematic"
    pps_brewer = "PPS Brewer"
    pps_hv = "PPS Hanurav-Vijayan"
    pps_murphy = "PPS Murphy"
    pps_rs = "PPS Rao-Sampford"
    pps_sys = "PPS Systematic"
    pps_wr = "PPS with replacement"
    grs = "General"


@unique
class SinglePSUEst(Enum):
    """Estimation options for strata with singleton PSU"""

    error = "Raise Error when one PSU in a stratum"
    skip = "Set variance to zero and skip stratum with one PSU"
    certainty = "Use SSUs or lowest units to estimate the variance"
    combine = "Combine the strata with the singleton psu to another stratum"



from __future__ import annotations

from typing import Optional, Union

import numpy as np

# import pandas as pd
import statsmodels.api as sm

from samplics.utils.formats import (
    fpc_as_dict,
    numpy_array,
)
from samplics.utils.types import Array, Number, Series, StringNumber


class SurveyGLM:
    """General linear models under complex survey sampling"""

    def __init__(self):
        self.beta: np.ndarray

    @staticmethod
    def _residuals(
        e: np.ndarray, psu: np.ndarray, nb_vars: Number
    ) -> tuple(np.ndarray, Number):

        psus = np.unique(psu)
        if psus.shape[0] == 1 and e.shape[0] == 1:
            raise AssertionError("Only one observation in the stratum")
        if psus.shape[0] == 1:
            psu = np.arange(e.shape[0])
            psus = np.unique(psu)
        e_values = np.zeros((psus.shape[0], nb_vars))

        for i, p in enumerate(np.unique(psus)):
            e_values[i, :] += np.sum(e[psu == p, :], axis=0)
        e_means = np.sum(e_values, axis=0) / psus.shape[0]

        return np.transpose(e_values - e_means) @ (e_values - e_means), psus.shape[0]

    def _calculate_g(
        self,
        samp_weight: np.ndarray,
        resid: np.ndarray,
        x: np.ndarray,
        stratum: np.ndarray,
        psu: np.ndarray,
        fpc: Union[dict[StringNumber, Number], Number],
        glm_scale=Number,
    ) -> np.ndarray:

        e = (samp_weight * resid)[:, None] * x / glm_scale
        if psu.shape in ((), (0,)):
            psu = np.arange(e.shape[0])
        if stratum.shape in ((), (0,)):
            e_h, n_h = self._residuals(e=e, psu=psu, nb_vars=x.shape[1])
            return fpc * (n_h / (n_h - 1)) * e_h
        else:
            g_h = np.zeros((x.shape[1], x.shape[1]))
            for s in np.unique(stratum):
                e_s = e[stratum == s, :]
                psu_s = psu[stratum == s]
                e_h, n_h = self._residuals(e=e_s, psu=psu_s, nb_vars=x.shape[1])
                g_h += fpc[s] * (n_h / (n_h - 1)) * e_h
            return g_h

    def estimate(
        self,
        y: Array,
        x: Optional[Array] = None,
        samp_weight: Optional[Array] = None,
        stratum: Optional[Series] = None,
        psu: Optional[Series] = None,
        fpc: Union[dict[StringNumber, Number], Series, Number] = 1.0,
        remove_nan: bool = False,
    ) -> None:

        _y = numpy_array(y)
        _x = numpy_array(x)
        _psu = numpy_array(psu)
        _stratum = numpy_array(stratum)
        _samp_weight = numpy_array(samp_weight)

        if _samp_weight.shape in ((), (0,)):
            _samp_weight = np.ones(y.shape[0])
        if _samp_weight.shape[0] == 1:
            _samp_weight = _samp_weight * np.ones(_y.shape[0])

        if not isinstance(fpc, dict):
            self.fpc = fpc_as_dict(_stratum, fpc)
        else:
            if np.unique(_stratum).tolist() != list(fpc.keys()):
                raise AssertionError(
                    "fpc dictionary keys must be the same as the strata!"
                )
            else:
                self.fpc = fpc

        glm_model = sm.GLM(endog=_y, exog=_x, var_weights=_samp_weight)
        glm_results = glm_model.fit()

        g = self._calculate_g(
            samp_weight=_samp_weight,
            resid=glm_results.resid_response,
            x=_x,
            stratum=_stratum,
            psu=_psu,
            fpc=self.fpc,
            glm_scale=glm_results.scale,
        )

        d = glm_results.cov_params()

        self.beta = glm_results.params
        self.cov_beta = (d @ g) @ d



"""EBLUP Area Model

The module implements the basic EBLUP area level model initially developed by Fay, R.E. and
Herriot, R.A. (1979) [#fh1979]_ and explicitly (mathematically) formulated in Rao, J.N.K. and
Molina, I. (2015) [#rm2015]_. The functionalities are organized in one class *EblupAreaModel* with
two main methods: *fit()* and *predict()*. The unit level standard error is assume known and the
model only estimate the fixed effects and the standard error of the random effects. The model
parameters can be fitted using restricted maximum likelihood (REML), maximum likelihood (ML) or
Fay-Herriot (FH). The method *predict()* provides both the mean and the MSE estimates at the area
level.

.. [#fh1979] Fay, R.E. and Herriot, R.A. (1979), Estimation of Income from Small Places: An
   Application of James-Stein Procedures to Census Data. *Journal of the American Statistical
   Association*, **74**, 269-277.
"""

from __future__ import annotations

import math

from typing import Optional, Union

import numpy as np
import pandas as pd

from samplics.utils.formats import dict_to_dataframe, numpy_array
from samplics.utils.types import Array, DictStrNum, Number, FitMethod


class EblupAreaModel:
    """*EblupAreaModel* implements the basic area level model.

    *EblupAreaModel* takes the unstable survey estimates as input and model them to produce the
    small area estimates. The standard error of the unit level residuals are assumed known and
    usually correspond to the survey estimates standard error. In some cases, the survey standard
    errors (or variances) are modelled to reduce their variability and the modelled standard errors
    are them used for the computes the small area estimates. The user can pick between REML, ML and
    FH to estimate the model parameters. EblupAreaModel provides both the area level point estimates and the taylor approximation of the MSEs.

    Setting attributes
        | method (str): the fitting method of the model parameters which can take the possible
        |   values restricted maximum likelihood (REML), maximum likelihood (ML), and
        |   Fay-Herriot (FH). If not specified, FitMethod.reml is used as default.

    Sample related attributes
        | yhat (array): the survey output area level estimates. This is also referred to as the
        | direct estimates.
        | error_std (number): the estimated standard error of the direct estimates.
        | X (ndarray): the auxiliary information.
        | b_const (array): an array of scaling parameters for the area random effects.
        | area (array): the area of the survey output estimates.
        | samp_size (dict): the sample size per small areas from the sample.
        | ys_mean (array): sample area means of the output variable.
        | Xs_mean (ndarray): sample area means of the auxiliary variables.

    Model fitting attributes
        | fitted (boolean): indicates whether the model has been fitted or not.
        | fixed_effects (array): the estimated fixed effects of the regression model.
        | fe_std (array): the estimated standard errors of the fixed effects.
        | random_effects (array): the estimated standard errors of the fixed effects.
        | re_std (number): the estimated area level standard error of the random effects.
        | convergence (dict): a dictionnary holding the convergence status and the number of
        |   iterations from the model fitting algorithm.
        | goodness (dict): a dictionary holding the log-likelihood, AIC, and BIC.

    Prediction related attributes
        | area_est (array): area level modelled estimates.
        | area_mse (array): area level taylor estimation of the MSE.

    Main methods
        | fit(): fits the linear mixed model to estimate the model parameters using REMl or ML
        |   methods.
        | predict(): predicts the area level mean estimates which includes both the point   | |
        |   estimates and the taylor MSE estimate.
    """

    def __init__(self, method: FitMethod = FitMethod.reml) -> None:
        if method not in (FitMethod.fh, FitMethod.ml, FitMethod.reml):
            raise AssertionError("Parameter method must be 'FH', 'ML, or 'REML'!")
        else:
            self.method = method

        # Sample data
        self.yhat: np.ndarray
        self.error_std: np.ndarray
        self.X: np.ndarray
        self.area: np.ndarray

        # Fitting stats
        self.fitted: bool = False
        self.fixed_effects: np.ndarray
        self.fe_std: np.ndarray
        self.re_std: Number
        self.convergence: dict[str, Union[float, int, bool]] = {}
        self.goodness: dict[str, Number] = {}  # loglikehood, deviance, AIC, BIC

        # Predict(ino/ed) data
        self.area_est: DictStrNum
        self.area_mse: DictStrNum
        self.area_mse_as1: DictStrNum
        self.area_mse_as2: DictStrNum
        self.area_mse_terms: dict[str, DictStrNum]

    def __str__(self) -> str:
        estimation = pd.DataFrame()
        estimation["area"] = self.area
        estimation["estimate"] = self.area_est
        estimation["mse"] = self.area_est

        fit = pd.DataFrame()
        fit["beta_coef"] = self.fixed_effects
        fit["beta_stderr"] = np.diag(self.fe_std)

        return f"""\n\nFH Area Model - Best predictor,\n\nConvergence status: {self.convergence['achieved']}\nNumber of iterations: {self.convergence['iterations']}\nPrecision: {self.convergence['precision']}\n\nGoodness of fit: {self.goodness}\n\nEstimation:\n{estimation}\n\nFixed effect:\n{fit}\n\nRandom effect variance:\n{self.re_std**2}\n\n"""

    def __repr__(self) -> str:
        return self.__str__()

    @staticmethod
    def _fixed_coefficients(
        area: np.ndarray,
        yhat: np.ndarray,
        X: np.ndarray,
        sigma2_e: np.ndarray,
        sigma2_v: float,
        b_const: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray]:
        # V = np.diag(sigma2_v * (b_const**2) + sigma2_e)
        # V_inv = np.linalg.inv(V)
        V_inv = np.diag(1 / (sigma2_v * (b_const**2) + sigma2_e))
        x_v_X_inv = np.linalg.inv(np.matmul(np.matmul(np.transpose(X), V_inv), X))
        x_v_x_inv_x = np.matmul(np.matmul(x_v_X_inv, np.transpose(X)), V_inv)
        beta_hat = np.matmul(x_v_x_inv_x, yhat)
        beta_cov = np.matmul(np.matmul(np.transpose(X), V_inv), X)

        return beta_hat.ravel(), np.linalg.inv(beta_cov)

    def _log_likelihood(
        self, y: np.ndarray, X: np.ndarray, beta: np.ndarray, V: np.ndarray
    ) -> Number:
        m = y.size
        const = m * np.log(2 * np.pi)
        ll_term1 = np.log(np.linalg.det(V))
        V_inv = np.linalg.inv(V)
        resid_term = y - np.dot(X, beta)
        if self.method in (FitMethod.ml, FitMethod.fh):  # Whta is likelihood for FH
            resid_var = np.dot(np.transpose(resid_term), V_inv)
            ll_term2 = np.dot(resid_var, resid_term)
            loglike = -0.5 * (const + ll_term1 + ll_term2)
        elif self.method == FitMethod.reml:
            xT_vinv_x = np.dot(np.dot(np.transpose(X), V_inv), X)
            ll_term2 = np.log(np.linalg.det(xT_vinv_x))
            ll_term3 = np.dot(np.dot(y, V_inv), resid_term)
            loglike = -0.5 * (const + ll_term1 + ll_term2 + ll_term3)
        else:
            raise AssertionError("A fitting method must be specified.")

        return float(loglike)

    def _partial_derivatives(
        self,
        area: np.ndarray,
        yhat: np.ndarray,
        X: np.ndarray,
        sigma2_e: np.ndarray,
        sigma2_v: Number,
        b_const: np.ndarray,
    ) -> tuple[Number, Number]:
        deriv_sigma = 0.0
        info_sigma = 0.0
        if self.method == FitMethod.ml:
            beta, beta_cov = self._fixed_coefficients(
                area=area,
                yhat=yhat,
                X=X,
                sigma2_e=sigma2_e,
                sigma2_v=sigma2_v,
                b_const=b_const,
            )
            for d in area:
                b_d = b_const[area == d]
                phi_d = sigma2_e[area == d]
                X_d = X[area == d, :]
                yhat_d = yhat[area == d]
                mu_d = np.matmul(X_d, beta)
                resid_d = yhat_d - mu_d
                sigma2_d = sigma2_v * (b_d**2) + phi_d
                term1 = (b_d**2 / sigma2_d)[0]
                term2 = (((b_d**2) * (resid_d**2)) / (sigma2_d**2))
                deriv_sigma += -0.5 * (term1 - term2)
                info_sigma += 0.5 * (term1**2)
        elif self.method == FitMethod.reml:
            B = np.diag(b_const**2)
            v_i = sigma2_e + sigma2_v * (b_const**2)
            V = np.diag(v_i)
            v_inv = np.linalg.inv(V)
            x_vinv_x = np.matmul(np.matmul(np.transpose(X), v_inv), X)
            x_xvinvx_x = np.matmul(np.matmul(X, np.linalg.inv(x_vinv_x)), np.transpose(X))
            P = v_inv - np.matmul(np.matmul(v_inv, x_xvinvx_x), v_inv)
            P_B = np.matmul(P, B)
            P_B_P = np.matmul(P_B, P)
            term1 = np.trace(P_B)
            term2 = np.matmul(np.matmul(np.transpose(yhat), P_B_P), yhat)
            deriv_sigma = -0.5 * (term1 - term2)
            info_sigma = 0.5 * np.trace(np.matmul(P_B_P, B))
        elif self.method == FitMethod.fh:  # Fay-Herriot approximation
            beta, beta_cov = self._fixed_coefficients(
                area=area,
                yhat=yhat,
                X=X,
                sigma2_e=sigma2_e,
                sigma2_v=sigma2_v,
                b_const=b_const,
            )
            for d in area:
                b_d = b_const[area == d]
                phi_d = sigma2_e[area == d]
                X_d = X[area == d, :]
                yhat_d = yhat[area == d]
                mu_d = np.dot(X_d, beta)
                resid_d = yhat_d - mu_d
                sigma2_d = sigma2_v * (b_d**2) + phi_d
                deriv_sigma += ((resid_d**2) / sigma2_d)[0]
                info_sigma += -(((b_d**2) * (resid_d**2)) / (sigma2_d**2))[0]
            m = yhat.size
            p = X.shape[1]
            deriv_sigma = m - p - deriv_sigma

        return deriv_sigma, info_sigma

    def _iterative_fisher_scoring(
        self,
        area: np.ndarray,
        yhat: np.ndarray,
        X: np.ndarray,
        sigma2_e: np.ndarray,
        b_const: np.ndarray,
        sigma2_v_start: float,
        tol: float,
        maxiter: int,
    ) -> tuple[float, float, int, float, bool]:  # May not need variance
        """Fisher-scroring algorithm for estimation of variance component
        return (sigma, covariance, number_iterations, tolerance, covergence status)"""

        iterations = 0
        tolerance = tol + 1.0
        sigma2_v = sigma2_v_start
        info_sigma = 0.0
        while iterations < maxiter and tolerance > tol:
            sigma2_v_previous = sigma2_v
            deriv_sigma, info_sigma = self._partial_derivatives(
                area=area,
                yhat=yhat,
                X=X,
                sigma2_e=sigma2_e,
                sigma2_v=sigma2_v,
                b_const=b_const,
            )
            sigma2_v += deriv_sigma / info_sigma
            tolerance = abs((sigma2_v - sigma2_v_previous) / sigma2_v_previous)
            iterations += 1

        return (
            max(sigma2_v, 0),
            1 / info_sigma,
            iterations,
            tolerance,
            tolerance <= tol,
        )

    def _eb_estimates(
        self,
        X: np.ndarray,
        beta: np.ndarray,
        area: np.ndarray,
        sigma2_e: np.ndarray,
        sigma2_v: Number,
        sigma2_v_cov: Number,
        b_const: np.ndarray,
    ) -> tuple[
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
    ]:
        m = self.yhat.size
        v_i = sigma2_e + sigma2_v * (b_const**2)
        V_inv = np.diag(1 / v_i)
        G = np.diag(np.ones(m) * sigma2_v)

        Z = np.diag(b_const)

        b = np.matmul(np.matmul(G, np.transpose(Z)), V_inv)
        d = np.transpose(X - np.matmul(np.transpose(b), X))
        x_vinv_x = np.matmul(np.matmul(np.transpose(X), V_inv), X)
        g2_term = np.linalg.inv(x_vinv_x)

        b_term_ml1 = np.linalg.inv(x_vinv_x)
        b_term_ml2_diag = (b_const**2) / (v_i**2)
        b_term_ml2 = np.matmul(np.matmul(np.transpose(X), np.diag(b_term_ml2_diag)), X)
        b_term_ml = float(np.trace(np.matmul(b_term_ml1, b_term_ml2)))

        estimates = np.array(self.yhat) * np.nan
        g1 = np.array(self.yhat) * np.nan
        g2 = np.array(self.yhat) * np.nan
        g3 = np.array(self.yhat) * np.nan
        g3_star = np.array(self.yhat) * np.nan

        g1_partial = np.array(self.yhat) * np.nan

        sum_inv_vi2 = np.sum(1 / (v_i**2))
        b_sigma2_v = 0.0
        if self.method == FitMethod.reml:
            g3_scale = 2.0 / sum_inv_vi2
        elif self.method == FitMethod.ml:
            b_sigma2_v = -(1.0 / 2.0 * sigma2_v_cov) * b_term_ml
            g3_scale = 2.0 / sum_inv_vi2
        elif self.method == FitMethod.fh:
            sum_vi = np.sum((1 / v_i))
            b_sigma2_v = 2.0 * (m * sum_inv_vi2 - sum_vi**2) / (sum_vi**3)
            g3_scale = 2.0 * m / sum_vi**2
        else:
            g3_scale = 0.0

        for d in area:
            b_d = b_const[area == d]
            phi_d = sigma2_e[area == d]
            X_d = X[area == d]
            yhat_d = self.yhat[area == d]
            mu_d = np.matmul(X_d, beta)
            resid_d = yhat_d - mu_d
            variance_d = sigma2_v * (b_d**2) + phi_d
            gamma_d = sigma2_v * (b_d**2) / variance_d
            estimates[area == d] = gamma_d * yhat_d + (1 - gamma_d) * mu_d
            g1[area == d] = gamma_d * phi_d
            g2_term_d = np.matmul(np.matmul(X_d, g2_term), np.transpose(X_d))
            g2[area == d] = ((1 - gamma_d) ** 2) * g2_term_d[0]
            g3[area == d] = ((1 - gamma_d) ** 2) * g3_scale / variance_d
            g3_star[area == d] = (g3[area == d] / variance_d) * (resid_d**2)
            g1_partial[area == d] = (b_d**2) * ((1 - gamma_d) ** 2) * b_sigma2_v

        mse = 0
        mse1_area_specific = 0
        mse2_area_specific = 0
        if self.method == FitMethod.reml:
            mse = g1 + g2 + 2 * g3
            mse1_area_specific = g1 + g2 + 2 * g3_star
            mse2_area_specific = g1 + g2 + g3 + g3_star
        elif self.method in (FitMethod.ml, FitMethod.fh):
            mse = g1 - g1_partial + g2 + 2 * g3
            mse1_area_specific = g1 - g1_partial + g2 + 2 * g3_star
            mse2_area_specific = g1 - g1_partial + g2 + g3 + g3_star

        return (
            np.asarray(estimates),
            np.asarray(mse),
            np.asarray(mse1_area_specific),
            np.asarray(mse2_area_specific),
            np.asarray(g1),
            np.asarray(g2),
            np.asarray(g3),
            np.asarray(g3_star),
        )

    def fit(
        self,
        yhat: Array,
        X: Array,
        area: Array,
        error_std: Array,
        re_std_start: float = 0.001,
        b_const: Union[np.ndarray, Number] = 1.0,
        intercept: bool = True,
        tol: float = 1e-4,
        maxiter: int = 100,
    ) -> None:
        """Fits the linear mixed models to estimate the fixed effects and the standard error of
        the random effects. In addition, the method provides statistics related to the model
        fitting e.g. convergence status, log-likelihood, and more.

        Args:
            yhat (Array): an array of the estimated area level survey estimates also called
            the direct estimates.
            X (Array): an multi-dimensional array of the auxiliary information associated to
            the sampled areas.
            area (Array): provides the areas associated to the direct estimates.
            error_std (Array): [description]
            re_std_start (float, optional): [description]. Defaults to 0.001.
            b_const (Union[np.array, Number], optional): [description]. Defaults to 1.0.
            tol (float, optional): tolerance used for convergence criteria. Defaults to 1.0e-4.
            maxiter (int, optional): maximum number of iterations for the fitting algorithm.
            Defaults to 100.
        """

        area = numpy_array(area)
        yhat = numpy_array(yhat)
        X = numpy_array(X)

        error_std = numpy_array(error_std)

        if (error_std <= 0).any():
            raise ValueError(
                "Some of the standard errors are not strictly positive. All standard errors must be greater than 0."
            )

        if isinstance(b_const, (int, float)):
            b_const = np.asarray(np.ones(area.size) * b_const)
        else:
            b_const = numpy_array(b_const)

        if intercept and isinstance(X, np.ndarray):
            X = np.insert(X, 0, 1, axis=1)

        (
            sigma2_v,
            sigma2_v_cov,
            iterations,
            tolerance,
            convergence,
        ) = self._iterative_fisher_scoring(
            area=area,
            yhat=yhat,
            X=X,
            sigma2_e=error_std**2,
            b_const=b_const,
            sigma2_v_start=re_std_start**2,
            tol=tol,
            maxiter=maxiter,
        )

        beta, beta_cov = self._fixed_coefficients(
            area=area,
            yhat=yhat,
            X=X,
            sigma2_e=error_std**2,
            sigma2_v=sigma2_v,
            b_const=b_const,
        )

        self.yhat = yhat
        self.error_std = error_std
        self.X = X
        self.area = area
        self.fixed_effects = beta
        self.fe_std = np.diag(beta_cov) ** (1 / 2)
        self.re_std = sigma2_v ** (1 / 2)
        self.re_std_cov = sigma2_v_cov

        self.convergence["achieved"] = convergence
        self.convergence["iterations"] = iterations
        self.convergence["precision"] = tolerance

        m = yhat.size
        p = X.shape[1] + 1
        Z_b2_Z = np.ones(shape=(m, m))
        V = np.diag(error_std**2) + sigma2_v * Z_b2_Z
        logllike = self._log_likelihood(yhat, X=X, beta=self.fixed_effects, V=V)
        self.goodness["loglike"] = logllike
        self.goodness["AIC"] = -2 * logllike + 2 * (p + 1)
        self.goodness["BIC"] = -2 * logllike + math.log(m) * (p + 1)

        self.fitted = True

    def predict(
        self,
        X: Array,
        area: Array,
        b_const: Union[np.ndarray, Number] = 1.0,
        intercept: bool = True,
    ) -> None:
        """Provides the modelled area levels estimates and their MSE estimates.

        Args:
            X (Array): an multi-dimensional array of the auxiliary variables associated to
            areas to predict.
            area (Array): provides the areas for the prediction.
            error_std (Array):
            b_const (Union[np.ndarray, Number], optional): [description]. Defaults to 1.0.

        Raises:
            Exception: [description]
        """

        area = numpy_array(area)
        if not self.fitted:
            raise Exception(
                "The model must be fitted first with .fit() before running the prediction."
            )

        X = numpy_array(X)
        if intercept and isinstance(X, np.ndarray):
            X = np.insert(X, 0, 1, axis=1)

        if isinstance(b_const, (int, float)):
            b_const = np.asarray(np.ones(area.size) * b_const)
        else:
            b_const = numpy_array(b_const)

        point_est, mse, mse1, mse2, g1, g2, g3, g3_star = self._eb_estimates(
            X=X,
            area=area,
            beta=self.fixed_effects,
            sigma2_e=self.error_std**2,
            sigma2_v=self.re_std**2,
            sigma2_v_cov=self.re_std_cov,
            b_const=b_const,
        )

        self.area_est = dict(zip(area, point_est))
        self.area_mse = dict(zip(area, mse))

    def to_dataframe(self, col_names: Optional(list) = None) -> pd.DataFrame:
        """Returns a pandas dataframe from dictionaries with same keys and one value per key.

        Args:
            col_names (list, optional): list of string to be used for the dataframe columns names.
                Defaults to ["_parameter", "_area", "_estimate", "_mse"].

        Returns:
            [type]: a pandas dataframe
        """

        est_df = dict_to_dataframe(col_names, self.area_est, self.area_mse)
        est_df.iloc[:, 0] = "mean"

        return est_df



"""Comparison module

The module implements comparisons of groups.

"""

from __future__ import annotations

import math

from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from scipy.stats import t

from samplics.estimation import TaylorEstimator
from samplics.utils.basic_functions import set_variables_names
from samplics.utils.checks import assert_probabilities
from samplics.utils.formats import numpy_array
from samplics.utils.types import Array, Number, Series, SinglePSUEst, StringNumber, PopParam


class Ttest:
    def __init__(
        self, samp_type: str, paired: bool = False, alpha: float = 0.05
    ) -> None:

        if samp_type.lower() not in ("one-sample", "two-sample"):
            raise ValueError(
                "Parameter 'type' must be equal to 'one-sample', 'two-sample'!"
            )
        assert_probabilities(x=alpha)

        self.samp_type = samp_type.lower()
        self.paired = paired

        self.point_est: Any = {}
        self.stats: Any = {}
        self.stderror: Any = {}
        self.stddev: Any = {}
        self.lower_ci: Any = {}
        self.upper_ci: Any = {}
        self.deff: Any = {}
        self.alpha: float = alpha
        # self.design_info: Dict[str, Number] = {}
        self.group_names: List[str] = []
        self.group_levels: Dict[str, StringNumber] = {}

    def __repr__(self) -> str:
        return f"T-test(samp_type={self.samp_type}, paired={self.paired}, alpha={self.alpha})"

    def __str__(self) -> str:

        if self.point_est == {}:
            return "No table to display"
        else:
            tbl_head = f"Design-based {self.samp_type.title()} T-test"
            if (
                self.samp_type == "one-sample" and self.group_names == []
            ) or self.paired:
                if self.samp_type == "one-sample":
                    tbl_subhead1 = (
                        f" Null hypothesis (Ho): mean = {self.stats['known_mean']}"
                    )
                else:
                    tbl_subhead1 = f" Null hypothesis (Ho): mean(Diff = {self.vars_names[0]} - {self.vars_names[1]}) = 0"
                tbl_subhead2 = f" t statistics: {self.stats['t']:.4f}"
                tbl_subhead3 = f" Degrees of freedom: {self.stats['df']:.2f}"
                tbl_subhead4 = " Alternative hypothesis (Ha):"
                tbl_subhead4a = (
                    f"  Prob(T < t) = {self.stats['p_value']['less_than']:.4f}"
                )
                tbl_subhead4b = (
                    f"  Prob(|T| > |t|) = {self.stats['p_value']['not_equal']:.4f}"
                )
                tbl_subhead4c = (
                    f"  Prob(T > t) = {self.stats['p_value']['greater_than']:.4f}"
                )

                return f"\n{tbl_head}\n{tbl_subhead1}\n{tbl_subhead2}\n{tbl_subhead3}\n{tbl_subhead4}\n{tbl_subhead4a}\n{tbl_subhead4b}\n{tbl_subhead4c} \n\n{self.to_dataframe().to_string(index=False)}\n"

            elif (self.samp_type == "one-sample" and self.group_names != []) or (
                self.samp_type == "two-sample" and not self.paired
            ):
                tbl_subhead1 = f" Null hypothesis (Ho): mean({self.group_names[0]}) = mean({self.group_names[1]}) "
                tbl_subhead2 = " Equal variance assumption:"
                tbl_subhead2a = f"  t statistics: {self.stats['t_eq_variance']:.4f}"
                tbl_subhead2b = (
                    f"  Degrees of freedom: {self.stats['df_eq_variance']:.2f}"
                )
                tbl_subhead3 = "  Alternative hypothesis (Ha):"
                tbl_subhead3a = f"   Prob(T < t) = {self.stats['p_value_eq_variance']['less_than']:.4f}"
                tbl_subhead3b = f"   Prob(|T| > |t|) = {self.stats['p_value_eq_variance']['not_equal']:.4f}"
                tbl_subhead3c = f"   Prob(T > t) = {self.stats['p_value_eq_variance']['greater_than']:.4f}"
                tbl_subhead4 = " Unequal variance assumption:"
                tbl_subhead4a = f"  t statistics: {self.stats['t_uneq_variance']:.4f}"
                tbl_subhead4b = (
                    f"  Degrees of freedom: {self.stats['df_uneq_variance']:.2f}"
                )
                tbl_subhead5 = "  Alternative hypothesis (Ha):"
                tbl_subhead5a = f"   Prob(T < t) = {self.stats['p_value_uneq_variance']['less_than']:.4f}"
                tbl_subhead5b = f"   Prob(|T| > |t|) = {self.stats['p_value_uneq_variance']['not_equal']:.4f}"
                tbl_subhead5c = f"   Prob(T > t) = {self.stats['p_value_uneq_variance']['greater_than']:.4f}"

                return f"\n{tbl_head}\n{tbl_subhead1}\n{tbl_subhead2}\n{tbl_subhead2a}\n{tbl_subhead2b}\n{tbl_subhead3}\n{tbl_subhead3a}\n{tbl_subhead3b}\n{tbl_subhead3c}\n{tbl_subhead4}\n{tbl_subhead4a}\n{tbl_subhead4b}\n{tbl_subhead5}\n{tbl_subhead5a}\n{tbl_subhead5b}\n{tbl_subhead5c} \n\n{self.to_dataframe().to_string(index=False)}\n"
            else:
                raise ValueError("Wrong specifications!")

    def _one_sample_one_group(
        self,
        y: np.ndarray,
        known_mean: Number,
        samp_weight: Array,
        stratum: Array,
        psu: Array,
        ssu: Array,
        fpc: Union[Dict, float] = 1,
        coef_var: bool = False,
        single_psu: Union[
            SinglePSUEst, dict[StringNumber, SinglePSUEst]
        ] = SinglePSUEst.error,
        strata_comb: Optional[dict[Array, Array]] = None,
    ) -> None:

        one_sample = TaylorEstimator(param=PopParam.mean, alpha=self.alpha)
        one_sample.estimate(
            y=y,
            samp_weight=samp_weight,
            stratum=stratum,
            psu=psu,
            ssu=ssu,
            fpc=fpc,
            coef_var=coef_var,
            single_psu=single_psu,
            strata_comb=strata_comb,
        )

        samp_mean = one_sample.point_est
        samp_std_dev = math.sqrt(y.shape[0]) * one_sample.stderror
        samp_t_value = math.sqrt(y.shape[0]) * (samp_mean - known_mean) / samp_std_dev
        left_p_value = t.cdf(samp_t_value, y.shape[0] - 1)

        # self.design_info = {
        #     "number_strata": one_sample.number_strata,
        #     "number_psus": one_sample.number_psus,
        #     "design_effect": 0,
        #     "degrees_of_freedom": one_sample.number_psus - one_sample.number_strata,
        # }

        self.stats = {
            "number_obs": y.shape[0],
            "known_mean": known_mean,
            "df": y.shape[0] - 1,
            "t": samp_t_value,
            "p_value": {
                "less_than": left_p_value,
                "greater_than": 1 - left_p_value,
                "not_equal": 2 * t.cdf(-abs(samp_t_value), y.shape[0] - 1),
            },
        }

        self.point_est = one_sample.point_est
        self.stderror = one_sample.stderror
        self.lower_ci = one_sample.lower_ci
        self.upper_ci = one_sample.upper_ci
        self.stddev = samp_std_dev

    def _two_groups_unpaired(
        self, mean_est: TaylorEstimator, group: np.ndarray
    ) -> Tuple[Any, Any, Any, Any, Any, Any]:

        if self.samp_type == "one-sample":
            group1 = mean_est.domains[0]
            group2 = mean_est.domains[1]
        elif self.samp_type == "two-sample":
            group1 = mean_est.by[0]
            group2 = mean_est.by[1]
        else:
            raise ValueError("Parameter type is not valid!")

        mean_group1 = mean_est.point_est[group1]
        mean_group2 = mean_est.point_est[group2]

        nb_obs_group1 = np.sum(group == group1)
        nb_obs_group2 = np.sum(group == group2)

        stddev_group1 = math.sqrt(nb_obs_group1) * mean_est.stderror[group1]
        stddev_group2 = math.sqrt(nb_obs_group2) * mean_est.stderror[group2]

        t_equal_variance = (mean_group1 - mean_group2) / (
            math.sqrt(
                (
                    (nb_obs_group1 - 1) * stddev_group1**2
                    + (nb_obs_group2 - 1) * stddev_group2**2
                )
                / (nb_obs_group1 + nb_obs_group2 - 2)
            )
            * math.sqrt(1 / nb_obs_group1 + 1 / nb_obs_group2)
        )

        t_df_equal_variance = nb_obs_group1 + nb_obs_group2 - 2

        t_unequal_variance = (mean_group1 - mean_group2) / math.sqrt(
            stddev_group1**2 / nb_obs_group1 + stddev_group2**2 / nb_obs_group2
        )

        t_df_unequal_variance = math.pow(
            stddev_group1**2 / nb_obs_group1 + stddev_group2**2 / nb_obs_group2, 2
        ) / (
            math.pow(stddev_group1**2 / nb_obs_group1, 2) / (nb_obs_group1 - 1)
            + math.pow(stddev_group2**2 / nb_obs_group2, 2) / (nb_obs_group2 - 1)
        )

        left_p_value_equal_variance = t.cdf(t_equal_variance, t_df_equal_variance)
        both_p_value_equal_variance = 2 * t.cdf(
            -abs(t_equal_variance), t_df_equal_variance
        )

        left_p_value_unequal_variance = t.cdf(t_unequal_variance, t_df_unequal_variance)
        both_p_value_unequal_variance = 2 * t.cdf(
            -abs(t_unequal_variance), t_df_unequal_variance
        )

        stats = {
            "number_obs": {group1: nb_obs_group1, group2: nb_obs_group2},
            "t_eq_variance": t_equal_variance,
            "df_eq_variance": t_df_equal_variance,
            "t_uneq_variance": t_unequal_variance,
            "df_uneq_variance": t_df_unequal_variance,
            "p_value_eq_variance": {
                "less_than": left_p_value_equal_variance,
                "greater_than": 1 - left_p_value_equal_variance,
                "not_equal": both_p_value_equal_variance,
            },
            "p_value_uneq_variance": {
                "less_than": left_p_value_unequal_variance,
                "greater_than": 1 - left_p_value_unequal_variance,
                "not_equal": both_p_value_unequal_variance,
            },
        }

        if (
            isinstance(mean_est.point_est, dict)
            and isinstance(mean_est.stderror, dict)
            and isinstance(mean_est.lower_ci, dict)
            and isinstance(mean_est.upper_ci, dict)
        ):
            point_est = mean_est.point_est.copy()
            stderror = mean_est.stderror.copy()
            lower_ci = mean_est.lower_ci.copy()
            upper_ci = mean_est.upper_ci.copy()
        else:
            point_est = mean_est.point_est
            stderror = mean_est.stderror
            lower_ci = mean_est.lower_ci
            upper_ci = mean_est.upper_ci

        stddev = {
            group1: math.sqrt(nb_obs_group1) * stderror[group1],
            group2: math.sqrt(nb_obs_group2) * stderror[group2],
        }

        return point_est, stderror, stddev, lower_ci, upper_ci, stats

    def _two_samples_unpaired(
        self,
        y: np.ndarray,
        group: np.ndarray,
        samp_weight: Optional[Array] = None,
        stratum: Optional[Array] = None,
        psu: Optional[Array] = None,
        ssu: Optional[Array] = None,
        fpc: Union[Dict, float] = 1,
        coef_var: bool = False,
        single_psu: Union[
            SinglePSUEst, dict[StringNumber, SinglePSUEst]
        ] = SinglePSUEst.error,
        strata_comb: Optional[dict[Array, Array]] = None,
    ) -> None:

        two_samples_unpaired = TaylorEstimator(param=PopParam.mean, alpha=self.alpha)
        two_samples_unpaired.estimate(
            y=y,
            by=group,
            samp_weight=samp_weight,
            stratum=stratum,
            psu=psu,
            ssu=ssu,
            fpc=fpc,
            coef_var=coef_var,
            single_psu=single_psu,
            strata_comb=strata_comb,
        )
        self._two_groups_unpaired(mean_est=two_samples_unpaired, group=group)

    def compare(
        self,
        y: Union[Series, Array],
        known_mean: Optional[Number] = None,
        group: Optional[Array] = None,
        varnames: Optional[Union[str, List[str]]] = None,
        samp_weight: Array = None,
        stratum: Optional[Array] = None,
        psu: Optional[Array] = None,
        ssu: Optional[Array] = None,
        fpc: Union[Dict, float] = 1,
        coef_var: bool = False,
        single_psu: Union[
            SinglePSUEst, dict[StringNumber, SinglePSUEst]
        ] = SinglePSUEst.error,
        strata_comb: Optional[dict[Array, Array]] = None,
        remove_nan: bool = False,
    ) -> None:

        if y is None:
            raise AssertionError("vars need to be an array-like object")
        if known_mean is None and group is None:
            raise AssertionError("Parameters 'known_mean' or 'group' must be provided!")
        if known_mean is not None and group is not None:
            raise AssertionError(
                "Only one parameter 'known_mean' or 'group' should be provided!"
            )

        if varnames is None:
            self.vars_names = set_variables_names(y, None, "var")
        elif isinstance(varnames, str):
            self.vars_names = set_variables_names(y, None, varnames)
        elif isinstance(varnames, list):
            self.vars_names = set_variables_names(y, varnames, varnames[0])
        else:
            raise AssertionError("varnames should be a string or a list of string")

        _y = numpy_array(y)
        _group = numpy_array(group)
        _samp_weight = numpy_array(samp_weight)
        _stratum = numpy_array(stratum)
        _psu = numpy_array(psu)
        _ssu = numpy_array(ssu)

        if self.samp_type == "one-sample" and known_mean is not None:
            self._one_sample_one_group(
                y=_y,
                known_mean=known_mean,
                samp_weight=_samp_weight,
                stratum=_stratum,
                psu=_psu,
                ssu=_ssu,
                fpc=fpc,
                coef_var=coef_var,
                single_psu=single_psu,
                strata_comb=strata_comb,
            )
        elif self.samp_type == "one-sample" and group is not None:
            one_sample = TaylorEstimator(param=PopParam.mean, alpha=self.alpha)
            one_sample.estimate(
                y=_y,
                domain=_group,
                samp_weight=_samp_weight,
                stratum=_stratum,
                psu=_psu,
                ssu=_ssu,
                fpc=fpc,
                coef_var=coef_var,
                single_psu=single_psu,
                strata_comb=strata_comb,
            )
            (
                self.point_est,
                self.stderror,
                self.stddev,
                self.lower_ci,
                self.upper_ci,
                self.stats,
            ) = self._two_groups_unpaired(mean_est=one_sample, group=_group)
            self.group_names = list(self.point_est.keys())
        elif self.samp_type == "two-sample" and not self.paired:
            self._two_samples_unpaired(
                y=_y,
                group=_group,
                samp_weight=_samp_weight,
                stratum=_stratum,
                psu=_psu,
                ssu=_ssu,
                fpc=fpc,
                coef_var=coef_var,
                single_psu=single_psu,
                strata_comb=strata_comb,
            )

            two_samples_unpaired = TaylorEstimator(param=PopParam.mean, alpha=self.alpha)
            two_samples_unpaired.estimate(
                y=_y,
                by=_group,
                samp_weight=_samp_weight,
                stratum=_stratum,
                psu=_psu,
                ssu=_ssu,
                fpc=fpc,
                coef_var=coef_var,
                single_psu=single_psu,
                strata_comb=strata_comb,
            )
            (
                self.point_est,
                self.stderror,
                self.stddev,
                self.lower_ci,
                self.upper_ci,
                self.stats,
            ) = self._two_groups_unpaired(mean_est=two_samples_unpaired, group=group)
            self.group_names = list(self.point_est.keys())
        elif self.samp_type == "two-sample" and self.paired:
            if len(_y.shape) == 1 or _y.shape[1] != 2:
                raise AssertionError(
                    "Parameter y must be an array-like object of dimension n by 2 for two-sample paired T-test"
                )
            d = _y[:, 0] - _y[:, 1]
            self._one_sample_one_group(
                y=d,
                known_mean=0,
                samp_weight=_samp_weight,
                stratum=_stratum,
                psu=_psu,
                ssu=_ssu,
                fpc=fpc,
            )
        else:
            raise ValueError("type must be equal to 'one-sample', 'two-sample'!")

    def to_dataframe(
        self,
    ) -> pd.DataFrame:

        if (self.samp_type == "one-sample" and self.group_names == []) or self.paired:
            return pd.DataFrame(
                data={
                    "Nb. Obs": [self.stats["number_obs"]],
                    PopParam.mean: [self.point_est],
                    "Std. Error": [self.stderror],
                    "Std. Dev.": [self.stddev],
                    "Lower CI": [self.lower_ci],
                    "Upper CI": [self.upper_ci],
                }
            )
        else:
            groups = list(self.point_est.keys())
            return pd.DataFrame(
                data={
                    "Group": groups,
                    "Nb. Obs": list(self.stats["number_obs"].values()),
                    PopParam.mean: list(self.point_est.values()),
                    "Std. Error": list(self.stderror.values()),
                    "Std. Dev.": list(self.stddev.values()),
                    "Lower CI": list(self.lower_ci.values()),
                    "Upper CI": list(self.upper_ci.values()),
                }
            )



from samplics.types.basic import (
    DF,
    Array,
    DictStrBool,
    DictStrFloat,
    DictStrInt,
    DictStrNum,
    Number,
    Series,
    StringNumber,
)
from samplics.types.containers import (
    # AuxVars,
    # DirectEst,
    # EbEst,
    # EbFit,
    # EblupEst,
    # EblupFit,
    # FitStats,
    SampleDesign,
)
from samplics.types.options import FitMethod, Mse, PopParam, SelectMethod, SinglePSUEst
from samplics.types.protocols import (
    DepVarsPrcl,
    FramePrcl,
    IndepVarsPrcl,
    SamplePrcl,
    ToDataFramePrcl,
)


__all__ = [
    "Array",
    # "AuxVars",
    "DepVarsPrcl",
    "DF",
    "DictStrNum",
    "DictStrInt",
    "DictStrFloat",
    "DictStrBool",
    # "DirectEst",
    # "EblupEst",
    # "EblupFit",
    "EbUnitModel",
    # "EbEst",
    # "EbFit",
    "FramePrcl",
    # "FitStats",
    "FitMethod",
    "IndepVarsPrcl",
    "Mse",
    "Number",
    "PopParam",
    "SampleDesign",
    "SamplePrcl",
    "SelectMethod",
    "Series",
    "SinglePSUEst",
    "StringNumber",
    "ToDataFramePrcl",
]



from samplics.datasets.datasets import (
    load_auto,
    load_birth,
    load_county_crop,
    load_county_crop_means,
    load_expenditure_milk,
    load_nhanes2,
    load_nhanes2brr,
    load_nhanes2jk,
    load_nmihs,
    load_psu_frame,
    load_psu_sample,
    load_ssu_sample,
)


__all__ = [
    "load_auto",
    "load_birth",
    "load_county_crop",
    "load_county_crop_means",
    "load_expenditure_milk",
    "load_nhanes2",
    "load_nhanes2brr",
    "load_nhanes2jk",
    "load_nmihs",
    "load_psu_frame",
    "load_psu_sample",
    "load_ssu_sample",
]



"""EBLUP for the unit level model.

This module implements the basic EBLUP unit level model. The functionalities are organized in
classes. Each class has three main methods: *fit()*, *predict()* and *bootstrap_mse()*.
Linear Mixed Models (LMM) are the core underlying statistical framework used to model the
hierarchical nature of the small area estimation (SAE) techniques implemented in this module,
see McCulloch, C.E. and Searle, S.R. (2001) [#ms2001]_ for more details on LMM.

The *EblupUnitModel* class implements the model developed by Battese, G.E., Harter, R.M., and
Fuller, W.A. (1988) [#bhf1988]_. The model parameters can fitted using restricted maximum
likelihood (REML) and maximum likelihood (ML). The normality assumption of the errors is not
necessary to predict the point estimates but is required for the taylor MSE estimation. The
predictions takes into account sampling rates. A bootstrap MSE estimation method is also implemted
for this class.

For a comprehensive review of the small area estimation models and its applications,
see Rao, J.N.K. and Molina, I. (2015) [#rm2015]_.

.. [#ms2001] McCulloch, C.E.and Searle, S.R. (2001), *Generalized, Linear, Mixed Models*,
   New York: John Wiley & Sons, Inc.
.. [#bhf1988] Battese, G.E., Harter, R.M., and Fuller, W.A. (1988). An error-components model for
   prediction of county crop areas using survey and satellite data, *Journal of the American
   Statistical Association*, **83**, 28-36.
.. [#rm2015] Rao, J.N.K. and Molina, I. (2015), *Small area estimation, 2nd edn.*,
   John Wiley & Sons, Hoboken, New Jersey.
"""

from __future__ import annotations

import warnings

from typing import Optional, Union

import numpy as np
import pandas as pd
import statsmodels.api as sm

from samplics.sae.sae_core_functions import area_stats
from samplics.utils.basic_functions import sumby
from samplics.utils.formats import dict_to_dataframe, numpy_array
from samplics.utils.types import Array, DictStrNum, Number, FitMethod


class EblupUnitModel:
    """*EblupUnitModel* implements the basic unit level model for means (a linear indicator).

    *EblupUnitModel* takes the sample data as input and fits the basic linear mixed model.
    The user can pick between restricted maximum likelihood (REML) or maximum likelihood (ML)
    to fit the model parameters. Also, EblupUnitModel predicts the areas means and provides
    the point and mean squared error (MSE) estimates of the empirical Bayes linear
    unbiased (EBLUP). User can also obtain the bootstrap mse estimates of the MSE.

    Setting attributes
        | method (str): the fitting method of the model parameters which can take the possible
        |   values restricted maximum likelihood (REML) or maximum likelihood (ML).
        |   If not specified, FitMethod.reml is used as default.

    Sample related attributes
        | ys (array): the output sample observations.
        | Xs (ndarray): the auxiliary information.
        | scales (array): an array of scaling parameters for the unit levels errors.
        | afactors (array): sum of the inverse squared of scale.
        | areas (array): the full vector of small areas from the sampled observations.
        | areas_list (array): the list of small areas from the sample data.
        | samp_size (dict): the sample size per small areas from the sample.
        | ys_mean (array): sample area means of the output variable.
        | Xs_mean (ndarray): sample area means of the auxiliary variables.}}}}}

    Model fitting attributes
        | fitted (boolean): indicates whether the model has been fitted or not.
        | fixed_effects (array): the estimated fixed effects of the regression model.
        | fe_std (array): the estimated standard errors of the fixed effects.
        | random_effects (array): the estimated area level random effects.
        | re_std (number): the estimated standard error of the random effects.
        | error_std (number): the estimated standard error of the unit level residuals.
        | convergence (dict): a dictionnary holding the convergence status and the number of
        |   iterations from the model fitting algorithm.
        | goodness (dict): a dictionary holding the log-likelihood, AIC, and BIC.
        | gamma (dict): ratio of the between-area variability (re_std**2) to the total
        |   variability (re_std**2 + error_std**2 / a_factor).

    Prediction related attributes
        | areap (array): the list of areas for the prediction.
        | Xmean (array): population means of the auxiliary variables.
        | number_reps (int): number of replicates for the bootstrap MSE estimation.
        | samp_rate (dict): sampling rates at the area level.
        | area_est (array): area level EBLUP estimates.
        | area_mse (array): area level taylor estimation of the MSE.
        | area_mse_boot (array): area level bootstrap estimation of the MSE.

    Main methods
        | fit(): fits the linear mixed model to estimate the model parameters using REMl or ML
        |   methods.
        | predict(): predicts the area level mean estimates which includes both the point
        |   estimates and the taylor MSE estimate.
        | bootstrap_mse(): computes the area level bootstrap MSE estimates of the mean.
    """

    def __init__(
        self,
        method: FitMethod = FitMethod.reml,
    ):
        # Setting
        self.method = method
        if self.method not in (FitMethod.reml, FitMethod.ml):
            raise AssertionError("Method must be 'REML' or 'ML'!")

        # Sample data
        self.scales: np.ndarray
        self.afactors: DictStrNum
        self.ys: np.ndarray
        self.Xs: np.ndarray
        self.areas: np.ndarray
        self.areas_list: np.ndarray
        self.samp_size: DictStrNum
        self.ys_mean: np.ndarray
        self.Xs_mean: np.ndarray

        # Fitting stats
        self.fitted: bool = False
        self.fixed_effects: np.ndarray
        self.fe_std: np.ndarray
        self.random_effects: np.ndarray
        self.re_std: Number = 0
        self.error_std: Number = 0
        self.convergence: dict[str, Union[float, int, bool]] = {}
        self.goodness: dict[str, Number] = {}  # loglikehood, deviance, AIC, BIC
        self.gamma: DictStrNum

        # Predict(ion/ed) data
        self.areap: np.ndarray
        self.Xp_mean: np.ndarray
        self.number_reps: int = 0
        self.samp_rate: DictStrNum
        self.area_est: DictStrNum
        self.area_mse: DictStrNum
        self.area_mse_boot: Optional[DictStrNum] = None

    def _beta(
        self,
        y: np.ndarray,
        X: np.ndarray,
        area: np.ndarray,
        weight: np.ndarray,
    ) -> np.ndarray:

        Xw = X * weight[:, None]
        p = X.shape[1]
        beta1 = np.zeros((p, p))
        beta2 = np.zeros(p)
        for d in np.unique(area):
            aread = area == d
            w_d = weight[aread]
            y_d = y[aread]
            X_d = X[aread]
            Xw_d = Xw[aread]
            Xw_d_bar = np.sum(Xw_d, axis=0) / np.sum(w_d)
            resid_d_w = X_d - Xw_d_bar * self.gamma[d]
            beta1 = beta1 + np.matmul(np.transpose(Xw_d), resid_d_w)
            beta2 = beta2 + np.sum(resid_d_w * y_d[:, None] * w_d[:, None], axis=0)

        return np.asarray(np.matmul(np.linalg.inv(beta1), beta2))

    def _mse(
        self,
        areas: np.ndarray,
        Xs_mean: np.ndarray,
        Xp_mean: np.ndarray,
        gamma: np.ndarray,
        samp_size: np.ndarray,
        afactor: np.ndarray,
        A_inv: np.ndarray,
    ) -> np.ndarray:

        sigma2e = self.error_std**2
        sigma2u = self.re_std**2

        g1 = gamma * sigma2e / afactor

        xbar_diff = Xp_mean - gamma[:, None] * Xs_mean
        g2_matrix = xbar_diff @ A_inv @ np.transpose(xbar_diff)
        g2 = np.diag(g2_matrix)

        alpha = sigma2e + afactor * sigma2u
        i_vv = 0.5 * sum((afactor / alpha) ** 2)
        i_ee = 0.5 * sum((samp_size - 1) / (sigma2e**2) + 1 / (alpha**2))
        i_ve = 0.5 * sum(afactor / (alpha**2))
        i_determinant = i_vv * i_ee - i_ve * i_ve

        g3_afactor = (1 / afactor**2) * (1 / (sigma2u + sigma2e / afactor) ** 3)
        g3 = (
            g3_afactor
            * (
                (sigma2e**2) * i_ee
                + (sigma2u**2) * i_vv
                - 2 * (sigma2e * sigma2u) * (-i_ve)
            )
            / i_determinant
        )

        return np.asarray(g1 + g2 + 2 * g3)

    def fit(
        self,
        ys: Array,
        Xs: Array,
        areas: Array,
        samp_weight: Optional[Array] = None,
        scales: Union[Array, Number] = 1,
        intercept: bool = True,
        tol: float = 1e-6,
        maxiter: int = 100,
    ) -> None:
        """Fits the linear mixed models to estimate the model parameters that is the fixed
        effects, the random effects standard error and the unit level residuals' standard error.
        In addition, the method provides statistics related to the model fitting e.g. convergence
        status, log-likelihood, AIC, BIC, and more.

        Args:
            ys (Array): An array of the output sample observations.
            Xs (Array): An multi-dimensional array of the auxiliary information.
            areas (Array): An array of the sampled area provided at the unit level.
            samp_weight (Optional[Array], optional): An array of the sample weights.
                Defaults to None.
            scales (Union[Array, Number], optional): The scale factor for the unit level errors.
                If a single number of provided, the same number will be applied to all observations. Defaults to 1.
            intercept (bool, optional): An boolean to indicate whether an intercept need to be
                added to X. Defaults to True.
            tol (float, optional): tolerance used for convergence criteria. Defaults to 1.0e-4.
            maxiter (int, optional): maximum number of iterations for the fitting algorithm.
            Defaults to 100.
        """

        areas = numpy_array(areas)
        ys = numpy_array(ys)
        Xs = numpy_array(Xs)

        self.ys = ys
        self.Xs = Xs
        self.areas = areas
        self.areas_list = np.unique(areas)

        if intercept:
            if Xs.ndim == 1:
                n = Xs.shape[0]
                Xs = np.insert(Xs.reshape(n, 1), 0, 1, axis=1)
            else:
                Xs = np.insert(Xs, 0, 1, axis=1)
        if samp_weight is not None:
            samp_weight = numpy_array(samp_weight)

        if isinstance(scales, (float, int)):
            scales = np.asarray(np.ones(ys.shape[0]) * scales)
        else:
            scales = numpy_array(scales)
        self.scales = scales

        self.afactors = dict(zip(self.areas_list, sumby(areas, scales)))

        reml = True if self.method == FitMethod.reml else False
        basic_model = sm.MixedLM(ys, Xs, areas)
        fit_kwargs = {
            "tol": tol,
            "gtol": tol,
            # "pgtol": tol,
            "maxiter": maxiter,
        }  # TODO: to improve in the future. Check: statsmodels.LikelihoodModel.fit()
        basic_fit = basic_model.fit(reml=reml, full_output=True, **fit_kwargs)


        self.error_std = basic_fit.scale**0.5
        self.fixed_effects = basic_fit.fe_params
        self.fe_std = basic_fit.bse_fe
        self.re_std = basic_fit.cov_re[0][0] ** 0.5
        self.convergence["achieved"] = basic_fit.converged
        self.convergence["iterations"] = len(basic_fit.hist[0]["allvecs"]) - 1

        nb_obs = ys.shape[0]
        nb_variance_params = basic_fit.cov_re.shape[0] + 1
        if self.method == FitMethod.reml:  # page 111 - Rao and Molina (2015)
            aic = -2 * basic_fit.llf + 2 * nb_variance_params
            bic = (
                -2 * basic_fit.llf
                + np.log(nb_obs - self.fixed_effects.shape[0]) * nb_variance_params
            )
        elif self.method == FitMethod.ml:
            aic = -2 * basic_fit.llf + 2 * (
                self.fixed_effects.shape[0] + nb_variance_params
            )
            bic = -2 * basic_fit.llf + np.log(nb_obs) * (
                self.fixed_effects.shape[0] + nb_variance_params
            )
        else:
            aic = np.nan
            bic = np.nan
        self.goodness["loglike"] = basic_fit.llf
        self.goodness["AIC"] = aic
        self.goodness["BIC"] = bic
        self.ys_mean, Xs_mean, gamma, samp_size = area_stats(
            ys,
            Xs,
            areas,
            self.error_std,
            self.re_std,
            self.afactors,
            samp_weight,
        )
        self.random_effects = gamma * (self.ys_mean - Xs_mean @ self.fixed_effects)
        self.gamma = dict(zip(self.areas_list, gamma))
        self.samp_size = dict(zip(self.areas_list, samp_size))
        self.Xs_mean = Xs_mean[:, 1:]

        self.fitted = True

    def predict(
        self,
        Xmean: Array,
        area: Array,
        pop_size: Optional[Array] = None,
        intercept: bool = True,
    ) -> None:
        """Predicts the area level means and provides the taylor MSE estimation of the estimated
        area means.

        Args:
            Xmean (Array): a multi-dimensional array of the population means of the auxiliary
                variables.
            area (Array): An array of the areas in the same order as Xmean and Popsize.
            pop_size (Optional[Array], optional): An array of the population size for the same
                areas as in Xmean and area and in the same order. Defaults to None.
            intercept (bool, optional): [description]. Defaults to True.

        Raises:
            Exception: when predict() is called before fitting the model.
        """

        if not self.fitted:
            raise Exception(
                "The model must be fitted first with .fit() before running the prediction."
            )

        Xmean = numpy_array(Xmean)
        self.Xp_mean = Xmean
        if intercept:
            if Xmean.ndim == 1:
                n = Xmean.shape[0]
                Xp_mean = np.insert(Xmean.reshape(n, 1), 0, 1, axis=1)
                Xs_mean = np.insert(self.Xs_mean.reshape(n, 1), 0, 1, axis=1)
                Xs = np.insert(self.Xs.reshape(n, 1), 0, 1, axis=1)
            else:
                Xp_mean = np.insert(Xmean, 0, 1, axis=1)
                Xs_mean = np.insert(self.Xs_mean, 0, 1, axis=1)
                Xs = np.insert(self.Xs, 0, 1, axis=1)
        else:
            Xp_mean = self.Xp_mean
            Xs_mean = self.Xs_mean
            Xs = self.Xs

        area = numpy_array(area)
        pop_size_dict = {}
        if pop_size is not None:
            pop_size = numpy_array(pop_size)
            pop_size_dict = dict(zip(area, pop_size))

        self.areap = area

        ps = np.isin(self.areap, self.areas_list)

        mu = dict(zip(self.areap, Xp_mean @ self.fixed_effects))
        resid = dict(zip(self.areas_list, self.ys_mean - Xs_mean @ self.fixed_effects))

        area_est = {}
        samp_rate = {}
        for d in self.areap:
            if pop_size is not None and d in self.areas_list:
                samp_rate[d] = self.samp_size[d] / pop_size_dict[d]
            else:
                samp_rate[d] = 0
            if d in self.areas_list:
                area_est[d] = (
                    mu[d]
                    + (samp_rate[d] + (1 - samp_rate[d]) * self.gamma[d]) * resid[d]
                )
            else:
                area_est[d] = mu[d]

        self.samp_rate = samp_rate
        self.area_est = area_est

        A_ps = (
            np.diag(np.zeros(Xp_mean.shape[1]))
            if Xp_mean.ndim >= 2
            else np.asarray([0])
        )

        ps_area_list = self.areap[ps]
        for d in ps_area_list:
            areadps = self.areas == d
            n_ps_d = np.sum(areadps)
            X_ps_d = Xs[areadps]
            scale_ps_d = self.scales[areadps]
            V_ps_d = (self.error_std**2) * np.diag(scale_ps_d) + (
                self.re_std**2
            ) * np.ones([n_ps_d, n_ps_d])
            A_ps = A_ps + np.transpose(X_ps_d) @ np.linalg.inv(V_ps_d) @ X_ps_d

        ps_area_indices = np.isin(self.areas_list, ps_area_list)
        a_factor_ps = np.asarray(list(self.afactors.values()))[ps_area_indices]
        gamma_ps = np.asarray(list(self.gamma.values()))[ps_area_indices]
        samp_size_ps = np.asarray(list(self.samp_size.values()))[ps_area_indices]
        mse_ps = self._mse(
            ps_area_list,
            Xs_mean[ps_area_indices],
            Xp_mean[ps],
            gamma_ps,
            samp_size_ps,
            a_factor_ps,
            np.linalg.inv(A_ps),
        )
        self.area_mse = dict(zip(ps_area_list, mse_ps))

        # TODO: add non-sampled areas prediction (section 7.2.2, Rao and molina (2015))
        # pr_area_list = self.areap[~ps]

    def bootstrap_mse(
        self,
        number_reps: int = 500,
        intercept: bool = True,
        tol: float = 1.0e-6,
        maxiter: int = 100,
        show_progress: bool = True,
    ) -> None:
        """Computes the MSE bootstrap estimates of the area level mean estimates.

        Args:
            number_reps (int): Number of replicates for the bootstrap method.. Defaults to 500.
            scale (Union[Array, Number], optional): [description]. Defaults to 1.
            intercept (bool, optional): [description]. Defaults to True.
            tol: float = 1.0e-4,
            maxiter: int = 100,
            show_progress (bool, optional): shows a bar progress of the bootstrap replicates
                calculations. Defaults to True.

        Raises:
            Exception: when bootstrap_mse() is called before fitting the model.
        """

        if not self.fitted:
            raise Exception(
                "The model must be fitted first with .fit() before running the prediction."
            )

        if intercept:
            if self.Xs.ndim == 1:
                n = self.Xs.shape[0]
                Xp_mean = np.insert(self.Xp_mean.reshape(n, 1), 0, 1, axis=1)
                Xs = np.insert(self.Xs.reshape(n, 1), 0, 1, axis=1)
            else:
                Xp_mean = np.insert(self.Xp_mean, 0, 1, axis=1)
                Xs = np.insert(self.Xs, 0, 1, axis=1)
        else:
            Xp_mean = self.Xp_mean
            Xs = self.Xs

        nb_areas = self.areap.size
        ps = np.isin(self.areap, self.areas_list)
        ps_area_list = self.areap[ps]

        for i, d in enumerate(ps_area_list):
            aread = self.areas == d
            scale_ps_d = self.scales[aread]
            error_d = np.abs(scale_ps_d) * np.random.normal(
                scale=self.error_std, size=(number_reps, self.samp_size[d])
            )
            re_boot_d = np.random.normal(scale=self.re_std, size=(number_reps, 1))
            re_d = np.repeat(re_boot_d, int(self.samp_size[d]), axis=1)

            if i == 0:
                error = error_d
                re_boot = re_boot_d
                re = re_d
                X_ps = Xs[aread]
                area_ps = self.areas[aread]
            else:
                error = np.append(error, error_d, axis=1)
                re_boot = np.append(re_boot, re_boot_d, axis=1)
                re = np.append(re, re_d, axis=1)
                X_ps = np.append(X_ps, Xs[aread], axis=0)
                area_ps = np.append(area_ps, self.areas[aread])

        mu = X_ps @ self.fixed_effects
        y_ps_boot = mu[None, :] + re + error
        i = 0
        bar_length = min(50, number_reps)
        steps = np.linspace(0, number_reps, bar_length).astype(int)

        fit_kwargs = {
            "tol": tol,
            "gtol": tol,
            # "pgtol": tol,
            "maxiter": maxiter,
        }  # TODO: to improve in the future. Check: statsmodels.LikelihoodModel.fit()
        reml = True if self.method == FitMethod.reml else False

        boot_mse = np.zeros((number_reps, nb_areas))
        print(f"Running the {number_reps} bootstrap iterations")
        for b in range(number_reps):
            with warnings.catch_warnings(record=True):
                boot_model = sm.MixedLM(y_ps_boot[b, :], X_ps, area_ps)
                boot_fit = boot_model.fit(reml=reml, **fit_kwargs)
            boot_fe = boot_fit.fe_params
            boot_error_std = boot_fit.scale**0.5
            boot_re_std = boot_fit.cov_re[0][0] ** 0.5
            boot_ys_mean, boot_Xs_mean, boot_gamma, _ = area_stats(
                y_ps_boot[b, :],
                X_ps,
                area_ps,
                boot_error_std,
                boot_re_std,
                self.afactors,
                None,
            )
            boot_re = boot_gamma * (boot_ys_mean - boot_Xs_mean @ boot_fe)
            boot_mu = Xp_mean @ self.fixed_effects + re_boot[b, :]
            boot_mu_h = Xp_mean @ boot_fe + boot_re
            boot_mse[b, :] = (boot_mu_h - boot_mu) ** 2

            if show_progress:
                if b in steps:
                    i += 1
                    print(
                        f"\r[%-{bar_length-1}s] %d%%"
                        % ("=" * i, 2 + (100 / bar_length) * i),
                        end="",
                    )
        if show_progress:
            print("\n")

        self.area_mse_boot = dict(
            zip(ps_area_list, np.asarray(np.mean(boot_mse, axis=0)))
        )

        # TODO: nonnegligeable sampling fractions, section 7.2.4, Rao and Molina (2015)

    def to_dataframe(
        self,
        col_names: Optional(list) = None,
        # col_names: list[str] = ["_parameter", "_area", "_estimate", "_mse", "_mse_boot"],
    ) -> pd.DataFrame:
        """Returns a pandas dataframe from dictionaries with same keys and one value per key.

        Args:
            col_names (list, optional): list of string to be used for the dataframe columns names.
                Defaults to ["_parameter", "_area", "_estimate", "_mse", "_mse_boot"].

        Returns:
            [pd.DataFrame]: a pandas dataframe
        """

        if self.area_est is None:
            raise AssertionError("No estimates yet. Must first run predict().")
        elif col_names is None:
            col_names = ["_parameter", "_area", "_estimate", "_mse", "_mse_boot"]
            if self.area_mse_boot is None:
                col_names.pop()
        else:
            ncols = len(col_names)
            if ncols != 5 and self.area_mse_boot is not None:
                raise AssertionError("col_names must have 5 values")

        if self.area_mse_boot is not None:
            est_df = dict_to_dataframe(
                col_names,
                self.area_est,
                self.area_mse,
                self.area_mse_boot,
            )
        else:
            est_df = dict_to_dataframe(
                col_names,
                self.area_est,
                self.area_mse,
            )
        est_df.iloc[:, 0] = "mean"

        return est_df



class SampleDesign:
    pass


# @frozen
# class AuxVars:
#     # TODO: Add missing values functionality
#     x: dict
#     nrecords: dict | Number
#     record_id: dict | None
#     domains: list | None
#     uid: int = int(dt.datetime.now(tz=dt.timezone.utc).strftime("%Y%m%d%H%M%S") + str(int(1e16 * rand.random())))

#     def __init__(
#         self,
#         x: DF | Array | Iterable[DF | Array] | None = None,
#         domain: Array | None = None,
#         record_id: Array = None,
#         **kwargs,
#     ) -> None:
#         assert x is not None or kwargs != {}

#         if isinstance(x, (DF, Array)):
#             aux_df = self.__from_df(x)
#             if isinstance(x, Array):
#                 aux_df.columns = ["__x_" + str(i) for i in range(aux_df.shape[1])]
#         elif isinstance(x, Iterable):
#             for i, d in enumerate(x):
#                 assert isinstance(d, (DF, Array))
#                 if i == 0:
#                     aux_df = self.__from_df(d)
#                     if isinstance(d, Array):
#                         aux_df.columns = ["__x_" + str(i) for i in range(aux_df.shape[1])]
#                 else:
#                     d_df = self.__from_df(d)
#                     if isinstance(d, Array):
#                         d_df.columns = ["__x_" + str(i) for i in range(d_df.shape[1])]
#                     aux_df.hstack([d_df], in_place=True)
#         else:
#             aux_df = None

#         if aux_df is None:
#             x = pl.from_dict(kwargs)
#         else:
#             x = aux_df.hstack(pl.from_dict(kwargs))

#         __record_id = (
#             numpy_array(record_id) if record_id is not None else np.linspace(0, x.shape[0] - 1, x.shape[0]).astype(int)
#         )

#         __domains = None
#         if domain is not None:
#             __domain = numpy_array(domain).tolist()
#             auxdata_dict = x.insert_column(0, pl.Series(__domain).alias("__domain")).partition_by(
#                 "__domain", as_dict=True
#             )

#             record_id_dict = pl.DataFrame([__domain, __record_id], schema=["__domain", "__record_id"]).partition_by(
#                 "__domain", as_dict=True
#             )

#             nrecords = {}
#             __domains = np.unique(__domain).tolist()
#             for d in __domains:
#                 auxdata_dict[d] = auxdata_dict[d].drop("__domain")
#                 record_id_dict[d] = record_id_dict[d]["__record_id"].to_list()
#                 nrecords[d] = auxdata_dict[d].shape[0]
#             auxdata_dict = {k: auxdata_dict[k].to_dict(as_series=False) for k in auxdata_dict}
#         else:
#             nrecords = x.shape[0]
#             auxdata_dict = x.to_dict()
#             record_id_dict = pl.DataFrame([__record_id], schema=["__record_id"]).to_dict()

#         # record_id_dict = {
#         #     k: record_id_dict[k].to_dict(as_series=False) for k in record_id_dict
#         # }

#         self.__attrs_init__(auxdata_dict, nrecords, record_id_dict, __domains)

#     def __from_df(self, auxdata: DF | Array) -> pl.DataFrame | None:
#         if isinstance(auxdata, pl.DataFrame):
#             return auxdata
#         elif isinstance(auxdata, pl.Series):
#             return pl.DataFrame(auxdata)
#         elif isinstance(auxdata, (pd.DataFrame, pd.Series)):
#             return pl.from_pandas(auxdata)
#         elif isinstance(auxdata, np.ndarray):
#             return pl.from_numpy(auxdata)
#         elif isinstance(auxdata, Array):
#             return pl.DataFrame(auxdata)
#         else:
#             return None

#     def to_numpy(
#         self,
#         keep_vars: str | Iterable[str] | None = None,
#         drop_vars: str | Iterable[str] | None = None,
#     ):
#         return self.to_polars(keep_vars=keep_vars, drop_vars=drop_vars).to_numpy()

#     def to_polars(
#         self,
#         keep_vars: str | Iterable[str] | None = None,
#         drop_vars: str | Iterable[str] | None = None,
#     ):
#         if self.domains is None:
#             auxdata = pl.from_dict(self.x).insert_at_idx(0, pl.Series(self.record_id).alias("__record_id"))
#         else:
#             auxdata = pl.concat(
#                 [
#                     pl.from_dict(self.x[d])
#                     .insert_at_idx(0, pl.Series(self.record_id[d]).alias("__record_id"))
#                     .insert_at_idx(
#                         1,
#                         pl.repeat(d, n=self.nrecords[d], eager=True).alias("__domain"),
#                     )
#                     for d in self.domains
#                 ]
#             )

#         if keep_vars is None and drop_vars is None:
#             varlist = auxdata.columns
#         elif keep_vars is not None:
#             varlist = keep_vars
#         elif drop_vars is not None:
#             varlist = [item for item in auxdata.columns if item not in drop_vars]
#         else:
#             raise TypeError("Keep_vars and drop_vars must be None or str or list[str]")

#         return auxdata.select(varlist)

#     def to_pandas(
#         self,
#         keep_vars: str | Iterable[str] | None = None,
#         drop_vars: str | Iterable[str] | None = None,
#     ):
#         return self.to_polars(keep_vars=keep_vars, drop_vars=drop_vars).to_pandas()


# class CovMat:
#     pass


# def _is_all_items_positive(obj: Array | DictStrNum) -> bool:
#     assert isinstance(obj, (np.ndarray, pd.Series, list, tuple, dict))
#     if isinstance(obj, dict):
#         arr = numpy_array(list(obj.values()))
#     else:
#         arr = numpy_array(obj).flatten()

#     if (arr <= 0).any():
#         return False
#     else:
#         return True

#     return (arr > 0).all()


# @frozen
# class DirectEst:
#     est: dict = field(validator=validators.instance_of(dict))
#     stderr: dict = field(validator=validators.instance_of(dict))
#     ssize: dict | None = None
#     psize: dict = None
#     domains: list | None = None
#     uid: int = int(dt.datetime.now(tz=dt.timezone.utc).strftime("%Y%m%d%H%M%S") + str(int(1e16 * rand.random())))

#     def __init__(
#         self,
#         est: dict | Array | Number,
#         stderr: dict | Array | Number,
#         ssize: dict | Array | Number | None = None,
#         psize: dict | Array | Number | None = None,
#         domain: Array | None = None,
#     ) -> None:
#         domains = None
#         if domain is not None:
#             domain = numpy_array(domain)
#             domains = domain.tolist()

#             if isinstance(stderr, Number):
#                 stderr = dict(zip(domain, tuple(np.repeat(stderr, len(domain)))))
#             if isinstance(stderr, Array):
#                 stderr = dict(zip(domain, stderr))
#             if isinstance(est, Number):
#                 est = dict(zip(domain, tuple(np.repeat(est, len(domain)))))
#             if isinstance(est, Array):
#                 est = dict(zip(domain, est))
#             if isinstance(ssize, Number):
#                 ssize = dict(zip(domain, tuple(np.repeat(ssize, len(domain)))))
#             if isinstance(ssize, Array):
#                 ssize = dict(zip(domain, ssize))

#         assert _is_all_items_positive(stderr)

#         if psize is not None:
#             if isinstance(ssize, Number):
#                 psize = dict(zip(domain, tuple(np.repeat(ssize, len(domain)))))
#             if isinstance(ssize, Array):
#                 psize = dict(zip(domain, stderr))

#         self.__attrs_init__(est, stderr, ssize, psize, domains)

#     @property
#     def cv(self):
#         return {
#             key: self.stderr[key] / self.est[key] if self.est[key] != 0 else float("inf") * self.stderr[key]
#             for key in self.stderr
#         }

#     def to_numpy(
#         self,
#         keep_vars: str | Iterable[str] | None = None,
#         drop_vars: str | Iterable[str] | None = None,
#     ):
#         return self.to_polars(keep_vars=keep_vars, drop_vars=drop_vars).to_numpy()

#     def to_polars(
#         self,
#         keep_vars: str | Iterable[str] | None = None,
#         drop_vars: str | Iterable[str] | None = None,
#     ):
#         if self.domains is None:
#             aux_df = pl.from_dict(
#                 {
#                     "est": list(self.est.values()),
#                     "stderr": list(self.stderr.values()),
#                     "ssize": list(self.ssize.values()),
#                 }
#             )
#         else:
#             aux_df = pl.from_dict(
#                 {
#                     "__domain": list(self.est.keys()),
#                     "est": list(self.est.values()),
#                     "stderr": list(self.stderr.values()),
#                     "ssize": list(self.ssize.values()),
#                 }
#             )

#         if keep_vars is None and drop_vars is None:
#             varlist = aux_df.columns
#         elif keep_vars is not None:
#             varlist = keep_vars
#         elif drop_vars is not None:
#             varlist = [item for item in aux_df.columns if item not in drop_vars]
#         else:
#             raise TypeError("Keep_vars and drop_vars must be None or str or list[str]")

#         return aux_df.select(varlist)

#     def to_pandas(
#         self,
#         keep_vars: str | Iterable[str] | None = None,
#         drop_vars: str | Iterable[str] | None = None,
#     ):
#         return self.to_polars(keep_vars=keep_vars, drop_vars=drop_vars).to_pandas()


# @frozen
# class FitStats:  # MAYBE call this ModelStats or FitStats or ...
#     method: FitMethod
#     err_stderr: dict
#     fe_est: namedtuple  # fixed effects
#     fe_stderr: namedtuple
#     re_stderr: float
#     re_stderr_var: float
#     log_llike: float
#     convergence: dict
#     goodness: dict
#     uid: int = int(dt.datetime.now(tz=dt.timezone.utc).strftime("%Y%m%d%H%M%S")) + int(1e16 * rand.random())


# @frozen
# class EblupFit:  # MAYBE call this ModelStats or FitStats or ...
#     method: FitMethod
#     err_stderr: dict
#     fe_est: namedtuple  # fixed effects
#     fe_stderr: namedtuple
#     re_stderr: float
#     re_stderr_var: float
#     log_llike: float
#     convergence: dict
#     goodness: dict
#     uid: int = int(dt.datetime.now(tz=dt.timezone.utc).strftime("%Y%m%d%H%M%S")) + int(1e16 * rand.random())


# @frozen
# class EbFit:
#     method: FitMethod
#     err_stderr: float
#     fe_est: namedtuple  # fixed effects
#     fe_stderr: namedtuple
#     re_stderr: float
#     re_stderr_var: float
#     log_llike: float
#     convergence: dict
#     goodness: dict
#     uid: int = int(dt.datetime.now(tz=dt.timezone.utc).strftime("%Y%m%d%H%M%S")) + int(1e16 * rand.random())


# @frozen
# class EblupEst:
#     pred: dict
#     domains: list
#     fit_stats: EblupFit
#     mse: dict | None = None
#     mse_boot: dict | None = None
#     mse_jkn: dict | None = None
#     uid: int = int(dt.datetime.now(tz=dt.timezone.utc).strftime("%Y%m%d%H%M%S")) + int(1e16 * rand.random())

#     def __init__(
#         self,
#         pred: dict | Array,
#         fit_stats: EblupFit,
#         domain: Array | None = None,
#         mse: dict | Array | None = None,
#         mse_boot: dict | Array | None = None,
#         mse_jkn: dict | Array | None = None,
#     ) -> None:
#         if domain is not None:
#             domain = numpy_array(domain)
#             domains = domain.tolist()

#             if isinstance(pred, Array):
#                 pred = dict(zip(domain, pred))
#             if isinstance(mse, Array):
#                 dict(zip(domain, mse))
#             if isinstance(mse_boot, Array):
#                 dict(zip(domain, mse_boot))
#             if isinstance(mse_jkn, Array):
#                 dict(zip(domain, mse_jkn))
#         else:
#             domains = list(pred.keys())

#         self.__attrs_init__(pred, domains, fit_stats, mse, mse_boot, mse_jkn)

#     @property
#     def rse():
#         pass

#     @property
#     def cv():
#         pass

#     def to_numpy():
#         pass

#     def to_polars():
#         pass

#     def to_pandas():
#         pass


# @frozen
# class EbEst:
#     area: list
#     pred: dict
#     fit_stats: EbFit
#     mse: dict | None = None
#     mse_boot: dict | None = None
#     mse_jkn: dict | None = None
#     uid: int = int(dt.datetime.now(tz=dt.timezone.utc).strftime("%Y%m%d%H%M%S")) + int(1e16 * rand.random())

#     @property
#     def rse():
#         pass

#     @property
#     def cv():
#         pass

#     def to_numpy():
#         pass

#     def to_polars():
#         pass

#     def to_pandas():
#         pass



"""Replicate weights module.

The module has one main class called *ReplicateWeight* which implements three replication
techniques: Bootstrap, Jackknife, and balanced repeated replication (BRR). For reference,
users can consultant Efron, B. and Tibshirani, R.J. (1994) [#et1994]_, Valliant, R. and
Dever, J. A. (2018) [#vd2018]_ and Wolter, K.M. (2007) [#w2007]_ for more details.

.. [#et1994] Efron, B. and Tibshirani, R.J. (1994), *An Introduction to the Boostrap*,
   Chapman & Hall/CRC.
.. [#w2007] Wolter, K.M. (2007), *Introduction to Variance Estimate, 2nd edn.*,
   Springer-Verlag New York, Inc
"""

from __future__ import annotations

import math

from typing import Optional, Union

import numpy as np
import pandas as pd

from samplics.utils import checks, formats
from samplics.utils import hadamard as hdd
from samplics.utils.types import Array, Number, RepMethod


class ReplicateWeight:
    """*ReplicateWeight* implements Boostrap, Jackknife and BRR to derive replicate weights.
    When possible design weights should be used as the input weights for creating the replicate
    weights, hence the weight adjustments can be applied to the replicates.

    Attributes:
        | method (str): replicate method.
        | fay_coef (float): Fay coefficient when implementing BRR-Fay.
        | nb_reps (int): number of replicates.
        | rep_coefs (np.ndarray): coefficients associated to the replicates.
        | strat (bool): strat indicator.
        | nb_psus (int): number of primary sampling units.
        | nb_strata (int): number of strata.
        | rand_seed (int): random seed.


    Methods:
        | replicate(): computes the replicate weights.

    """

    def __init__(
        self,
        method: RepMethod,
        strat: bool = True,
        nb_reps: int = 500,
        fay_coef: float = 0.0,
        rand_seed: Optional[int] = None,
    ):

        self.method = method
        self.strat = strat
        if self.method == RepMethod.bootstrap:
            self.nb_reps = nb_reps
            self.rep_coefs = list((1 / nb_reps) * np.ones(nb_reps))
        elif self.method == RepMethod.brr:
            self.nb_reps = 0
            self.fay_coef = fay_coef

        self.nb_psus = 0
        self.nb_strata = 0
        self.rep_coefs = []
        self.degree_of_freedom = 0
        if rand_seed is not None:
            self.rand_seed = rand_seed
            np.random.seed(rand_seed)

    def _reps_to_dataframe(
        self, psus: pd.DataFrame, rep_data: np.ndarray, rep_prefix: str
    ) -> pd.DataFrame:

        rep_data = pd.DataFrame(rep_data)
        rep_data.reset_index(drop=True, inplace=True)
        rep_data.rename(columns=lambda x: rep_prefix + str(x + 1), inplace=True)
        psus.reset_index(drop=True, inplace=True)
        rep_data = pd.concat([psus, rep_data], axis=1)

        return rep_data

    def _rep_prefix(self, prefix: Optional[str]) -> str:

        if self.method == RepMethod.jackknife and prefix is None:
            rep_prefix = "_jk_wgt_"
        elif self.method == RepMethod.bootstrap and prefix is None:
            rep_prefix = "_boot_wgt_"
        elif self.method == RepMethod.brr and prefix is None:
            rep_prefix = "_brr_wgt_"
        elif self.method == RepMethod.brr and self.fay_coef > 0 and prefix is None:
            rep_prefix = "_fay_wgt_"
        elif prefix is None:
            rep_prefix = "_rep_wgt_"
        else:
            rep_prefix = prefix

        return rep_prefix

    def _degree_of_freedom(
        self,
        weight: np.ndarray,
        stratum: Optional[np.ndarray],
        psu: np.ndarray,
    ) -> None:

        stratum = formats.numpy_array(stratum)
        psu = formats.numpy_array(psu)

        if stratum.size <= 1:
            self.degree_of_freedom = np.unique(psu).size - 1
        elif psu.size > 1:
            self.degree_of_freedom = np.unique(psu).size - np.unique(stratum).size
        else:
            weight = formats.numpy_array(weight)
            self.degree_of_freedom = weight.size

    # Bootstrap methods
    @staticmethod
    def _boot_psus_replicates(
        nb_psus: int,
        nb_reps: int,
        samp_rate: Number = 0,
        size_gap: int = 1,
    ) -> np.ndarray:
        """Creates the bootstrap replicates structure"""

        if nb_psus <= size_gap:
            raise AssertionError("size_gap should be smaller than the number of units")

        sample_size = nb_psus - size_gap
        psu = np.arange(0, nb_psus)
        psu_boot = np.random.choice(psu, size=(nb_reps, sample_size))
        psu_replicates = np.zeros(shape=(nb_psus, nb_reps))
        for rep in np.arange(0, nb_reps):
            psu_ids, psus_counts = np.unique(psu_boot[rep, :], return_counts=True)
            psu_replicates[:, rep][psu_ids] = psus_counts

        ratio_sqrt = np.sqrt((1 - samp_rate) * sample_size / (nb_psus - 1))

        return np.asarray(
            1 - ratio_sqrt + ratio_sqrt * (nb_psus / sample_size) * psu_replicates
        )

    def _boot_replicates(
        self,
        psu: np.ndarray,
        stratum: Optional[np.ndarray],
        samp_rate: Number = 0,
        size_gap: int = 1,
    ) -> np.ndarray:

        if stratum is None:
            psu_ids = np.unique(psu)
            boot_coefs = self._boot_psus_replicates(
                psu_ids.size, self.nb_reps, samp_rate, size_gap
            )
        else:
            strata = np.unique(stratum)
            for k, s in enumerate(strata):
                psu_ids_s = np.unique(psu[stratum == s])
                nb_psus_s = psu_ids_s.size
                boot_coefs_s = self._boot_psus_replicates(
                    nb_psus_s, self.nb_reps, samp_rate, size_gap
                )
                if k == 0:
                    boot_coefs = boot_coefs_s
                else:
                    boot_coefs = np.concatenate((boot_coefs, boot_coefs_s), axis=0)

        return boot_coefs

    # BRR methods
    def _brr_nb_reps(
        self, psu: np.ndarray, stratum: Optional[np.ndarray] = None
    ) -> None:

        if stratum is None:
            self.nb_psus = np.unique(psu).size
            self.nb_strata = self.nb_psus // 2 + self.nb_psus % 2
        else:
            self.nb_psus = np.unique(np.array(list(zip(stratum, psu))), axis=0).shape[0]
            self.nb_strata = np.unique(stratum).size
            if 2 * self.nb_strata != self.nb_psus:
                raise AssertionError(
                    "Number of psus must be twice the number of strata!"
                )

        if self.nb_reps < self.nb_strata:
            self.nb_reps = self.nb_strata

        if self.nb_reps <= 28:
            if self.nb_reps % 4 != 0:
                self.nb_reps = 4 * (self.nb_reps // 4 + 1)
        else:
            nb_reps_log2 = int(math.log(self.nb_reps, 2))
            if math.pow(2, nb_reps_log2) != self.nb_reps:
                self.nb_reps = int(math.pow(2, nb_reps_log2))

    def _brr_replicates(
        self, psu: np.ndarray, stratum: Optional[np.ndarray]
    ) -> np.ndarray:
        """Creates the brr replicate structure"""

        if not (0 <= self.fay_coef < 1):
            raise ValueError(
                "The Fay coefficient must be greater or equal to 0 and lower than 1."
            )
        self._brr_nb_reps(psu, stratum)

        self.rep_coefs = list(
            (1 / (self.nb_reps * pow(1 - self.fay_coef, 2))) * np.ones(self.nb_reps)
        )

        brr_coefs = hdd.hadamard(self.nb_reps).astype(float)
        brr_coefs = brr_coefs[:, 1 : self.nb_strata + 1]
        brr_coefs = np.repeat(brr_coefs, 2, axis=1)
        for r in np.arange(self.nb_reps):
            for h in np.arange(self.nb_strata):
                start = 2 * h
                end = start + 2
                if brr_coefs[r, start] == 1.0:
                    brr_coefs[r, start:end] = [
                        self.fay_coef,
                        2 - self.fay_coef,
                    ]
                else:  # brr_coefs[r, 2 * h] == -1:
                    brr_coefs[r, start:end] = [
                        2 - self.fay_coef,
                        self.fay_coef,
                    ]

        return brr_coefs.T

    # Jackknife
    @staticmethod
    def _jkn_psus_replicates(nb_psus: int) -> np.ndarray:
        """Creates the jackknife delete-1 replicate structure"""

        jk_coefs = (nb_psus / (nb_psus - 1)) * (
            np.ones((nb_psus, nb_psus)) - np.identity(nb_psus)
        )

        return np.asarray(jk_coefs)

    def _jkn_replicates(
        self, psu: np.ndarray, stratum: Optional[np.ndarray]
    ) -> np.ndarray:

        self.rep_coefs = ((self.nb_reps - 1) / self.nb_reps) * np.ones(self.nb_reps)

        if stratum is None:
            psu_ids = np.unique(psu)
            jk_coefs = self._jkn_psus_replicates(psu_ids.size)
        else:
            strata = np.unique(stratum)
            jk_coefs = np.ones((self.nb_reps, self.nb_reps))
            start = end = 0
            for s in strata:
                psu_ids_s = np.unique(psu[stratum == s])
                nb_psus_s = psu_ids_s.size
                end = start + nb_psus_s
                jk_coefs[start:end, start:end] = self._jkn_psus_replicates(nb_psus_s)
                self.rep_coefs[start:end] = (nb_psus_s - 1) / nb_psus_s
                start = end

        self.rep_coefs = list(self.rep_coefs)

        return jk_coefs

    def replicate(
        self,
        samp_weight: Array,
        psu: Array,
        stratum: Optional[Array] = None,
        rep_coefs: Union[Array, Number] = False,
        rep_prefix: Optional[str] = None,
        psu_varname: str = "_psu",
        str_varname: str = "_stratum",
    ) -> pd.DataFrame:
        """Computes replicate sample weights.

        Args:
            samp_weight (Array): array of sample weights. To incorporate the weights adjustment
                in the replicate weights, first replicate the design sample weights then apply
                the adjustments to the replicates.
            psu (Array):
            stratum (Array, optional): array of the strata. Defaults to None.
            rep_coefs (Union[Array, Number], optional): coefficients associated to the replicates.
                Defaults to False.
            rep_prefix (str, optional): prefix to apply to the replicate weights names.
                Defaults to None.
            psu_varname (str, optional): name of the psu variable in the output dataframe.
                Defaults to "_psu".
            str_varname (str, optional): name of the stratum variable in the output dataframe.
                Defaults to "_stratum".

        Raises:
            AssertionError: raises an assertion error when stratum is None for a stratified design.
            AssertionError: raises an assertion error when the replication method is not valid.

        Returns:
            pd.DataFrame: a dataframe of the replicates sample weights.
        """

        samp_weight = formats.numpy_array(samp_weight)
        psu = formats.numpy_array(psu)
        if not self.strat:
            stratum = None
        else:
            stratum = formats.numpy_array(stratum)

        self._degree_of_freedom(samp_weight, stratum, psu)

        if self.strat and stratum is None:
            raise AssertionError("For a stratified design, stratum must be specified.")
        elif stratum is not None:
            stratum_psu = pd.DataFrame({str_varname: stratum, psu_varname: psu})
            stratum_psu.sort_values(by=str_varname, inplace=True)
            key = [str_varname, psu_varname]
        elif self.method == RepMethod.brr:
            _, str_index = np.unique(psu, return_index=True)
            checks.assert_brr_number_psus(str_index)
            psus = psu[np.sort(str_index)]
            strata = np.repeat(range(1, psus.size // 2 + 1), 2)
            stratum_psu = pd.DataFrame({str_varname: strata, psu_varname: psus})
            psu_pd = pd.DataFrame({psu_varname: psu})
            stratum_psu = pd.merge(
                psu_pd, stratum_psu, on=psu_varname, how="left", sort=False
            )
            stratum_psu = stratum_psu[[str_varname, psu_varname]]
            key = [str_varname, psu_varname]
        else:
            stratum_psu = pd.DataFrame({psu_varname: psu})
            key = [psu_varname]

        psus_ids = stratum_psu.drop_duplicates()

        if self.method == RepMethod.jackknife:
            self.nb_reps = psus_ids.shape[0]
            _rep_data = self._jkn_replicates(psu, stratum)
        elif self.method == RepMethod.bootstrap:
            _rep_data = self._boot_replicates(psu, stratum)
        elif self.method == RepMethod.brr:
            _rep_data = self._brr_replicates(psu, stratum)
            self.rep_coefs = list(
                (1 / self.nb_reps * pow(1 - self.fay_coef, 2)) * np.ones(self.nb_reps)
            )
        else:
            raise AssertionError(
                "Replication method not recognized. Possible options are: 'bootstrap', 'brr', and 'jackknife'"
            )

        rep_prefix = self._rep_prefix(rep_prefix)
        _rep_data = self._reps_to_dataframe(psus_ids, _rep_data, rep_prefix)

        samp_weight = pd.DataFrame({"_samp_weight": samp_weight})
        samp_weight.reset_index(drop=True, inplace=True)
        full_sample = pd.concat([stratum_psu, samp_weight], axis=1)
        full_sample = pd.merge(full_sample, _rep_data, on=key, how="left", sort=False)

        if not rep_coefs:
            rep_cols = [col for col in full_sample if col.startswith(rep_prefix)]
            full_sample[rep_cols] = full_sample[rep_cols].mul(
                samp_weight.values, axis=0
            )

        return full_sample

    def normalize(self, rep_weights: Union[np.ndarray, pd.DataFrame]) -> pd.DataFrame:
        pass

    def adjust(self, rep_weights: Union[np.ndarray, pd.DataFrame]) -> pd.DataFrame:
        pass

    def trim(self, rep_weights: Union[np.ndarray, pd.DataFrame]) -> pd.DataFrame:
        pass
